"""
Module containing a class for getting the BNGL NET file for a given PySB model.

Serves as a wrapper around :py:func:`pysb.bng.generate_network`, which
generates the BNGL for the model and then invokes BNG to generate the NET file.

For information on how to use the model exporters, see the documentation
for :py:mod:`pysb.export`.
"""

from pysb.bng import generate_network
from pysb.export import Exporter

class BngNetExporter(Exporter):
    """A class for generating the BNG NET file for a given PySB model.

    Inherits from :py:class:`pysb.export.Export`, which implements
    basic functionality for all exporters.
    """
    def export(self):
        """Generate the BNGL NET file for the PySB model associated with
        the exporter. A wrapper around :py:func:`pysb.bng.generate_network`.

        Returns
        -------
        string
            The NET file output for the model, generated by BNG.
        """
        net_str = ''
        if self.docstring:
            net_str += '# ' + self.docstring.replace('\n', '\n# ') + '\n'
        net_str += generate_network(self.model, append_stdout=True)
        return net_str




"""A simple three-species chemical kinetics system known as "Robertson's
example", as presented in:

H. H. Robertson, The solution of a set of reaction rate equations, in Numerical
Analysis: An Introduction, J. Walsh, ed., Academic Press, 1966, pp. 178-182.
"""

# This is a simple system often used to study stiffness in systems of
# differential equations.  It doesn't leverage the power of rules-based modeling
# or pysb, but it's a useful small model for purposes of experimentation.
#
# A brief report addressing issues of stiffness encountered in numerical
# integration of Robertson's example can be found here:
# http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.53.8603&rep=rep1&type=pdf
#
# The chemical model is as follows:
#
#      Reaction        Rate
#   ------------------------
#       A -> B         0.04
#      2B -> B + C     3.0e7
#   B + C -> A + C     1.0e4
#
# The resultant system of differential equations is:
#
# y1' = -0.04 * y1 + 1.0e4 * y2 * y3
# y2' =  0.04 * y1 - 1.0e4 * y2 * y3 - 3.0e7 * y2^2
# y3' =                                3.0e7 * y2^2
#
# If you run this script directly, it will generate the equations and print them
# in a form very closely resembling that given above.  See also run_robertson.py
# which integrates the system and plots the trajectories.


from __future__ import print_function
from pysb import *

Model()

Monomer('A')
Monomer('B')
Monomer('C')

#       A -> B         0.04
Rule('A_to_B', A() >> B(), Parameter('k1', 0.04))
#      2B -> B + C     3.0e7
Rule('BB_to_BC', B() + B() >> B() + C(), Parameter('k2', 3.0e7))
#   B + C -> A + C     1.0e4
Rule('BC_to_AC', B() + C() >> A() + C(), Parameter('k3', 1.0e4))

# The system is known to be stiff for initial values A=1, B=0, C=0
Initial(A(), Parameter('A_0', 1.0))
Initial(B(), Parameter('B_0', 0.0))
Initial(C(), Parameter('C_0', 0.0))

# Observe total amount of each monomer
Observable('A_total', A())
Observable('B_total', B())
Observable('C_total', C())


if __name__ == '__main__':
    from pysb.bng import generate_equations

    # This creates model.odes which contains the math
    generate_equations(model)

    # Build a symbol substitution mapping to help the math look like the
    # equations in the comment above, instead of showing the internal pysb
    # symbol names
    # ==========
    substitutions = {}
    # Map parameter symbols to their values
    substitutions.update((p.name, p.value) for p in model.parameters)
    # Map species variables sI to yI+1, e.g. s0 -> y1
    substitutions.update(('s%d' % i, 'y%d' % (i+1))
                         for i in range(len(model.odes)))

    print(__doc__, "\n", model, "\n")

    # Iterate over each equation
    for i, eq in enumerate(model.odes):
        # Perform the substitution using the sympy 'subs' method and the
        # mappings we built above
        eq_sub = eq.subs(substitutions)
        # Display the equation
        print('y%d\' = %s' % (i+1, eq_sub))

    print("""
NOTE: This model code is designed to be imported and programatically
manipulated, not executed directly. The above output is merely a
diagnostic aid. Please see run_robertson.py for example usage.""")



import pysb
import abc
from pysb.core import SelfExporter
from pysb.pattern import SpeciesPatternMatcher, ReactionPatternMatcher, \
    RulePatternMatcher


class ModelAssertionFailure(Exception):
    def __init__(self, assertion, model, message=None):
        self.assertion = assertion
        self.model = model
        self.message = message

    def __repr__(self):
        base_msg = '%s on model %s failed' % (self.assertion,
                                              self.model.name)
        if self.message:
            return base_msg + ': ' + str(self.message)
        else:
            return base_msg

    def __str__(self):
        return repr(self)


class TestSuite(object):
    """
    A suite of tests for checking properties of a model

    There are two modes of operation: building a test suite using add() and
    executing all the tests at once with check_all(), or executing tests
    immediately with check().

    Examples
    --------

    Create a test suite for the EARM 1.0 model:

    >>> from pysb.testing.modeltests import TestSuite, SpeciesExists, \
        SpeciesDoesNotExist
    >>> from pysb.bng import generate_equations
    >>> from pysb.examples.earm_1_0 import model
    >>> ts = TestSuite(model)

    Create variables for model components (not needed for models defined
    interactively):

    >>> AMito, mCytoC, mSmac, cSmac, L, CPARP = [model.monomers[m] for m in \
                                       ('AMito', 'mCytoC', 'mSmac', 'cSmac', \
                                        'L', 'CPARP')]

    Add some assertions:

    Check that AMito(b=1) % mSmac(b=1) exists in the species graph (note this
    doesn't guarantee the species will actually be producted/consumed/change
    in concentration; that depends on the rate constants):

    >>> ts.add(SpeciesExists(AMito(b=1) % mSmac(b=1)))

    This is the opposite check, that the complex above doesn't exist,
    which should of course fail:

    >>> ts.add(SpeciesDoesNotExist(AMito(b=1) % mSmac(b=1)))

    We can also specify that species matching a pattern should never exist
    in a model. For example, we shouldn't ever be producing unbound
    ligand in the EARM 1.0 model:

    >>> ts.add(SpeciesNeverProduct(L(b=None)))

    We could also have used SpeciesOnlyReactant. The difference is the
    latter checks for an appearance as a reactant, whereas
    SpeciesNeverProduct would pass whether the species appeared as a
    reactant or not.

    >>> ts.add(SpeciesOnlyReactant(L(b=None)))

    CPARP is an output in this model, so it should appear as a product but
    never as a reactant:

    >>> ts.add(SpeciesOnlyProduct(CPARP()))

    When we're ready, we can generate the reactions and check the assertions:

    >>> generate_equations(model)
    >>> ts.check_all()  # doctest:+ELLIPSIS
    SpeciesExists(AMito() % mSmac())...OK...
    SpeciesDoesNotExist(AMito() % mSmac())...FAIL...
      [AMito(b=1) % mSmac(b=1)]...
    SpeciesExists(AMito(b=1) % mCytoC(b=1))...OK...
    SpeciesNeverProduct(L(b=None))...OK...
    SpeciesOnlyProduct(CPARP())...OK...

    We can also execute any test immediately without adding it to the test
    suite (note that some tests require a reaction network to be generated):

    >>> ts.check(SpeciesExists(L(b=None)))
    True
    """
    _KNOWN_CACHES = {'species_pattern_matcher': SpeciesPatternMatcher,
                     'rule_pattern_matcher': RulePatternMatcher,
                     'reaction_pattern_matcher': ReactionPatternMatcher}
    _COL = {'OK': '\033[92m', 'FAIL': '\033[91m', 'END': '\033[0m'}

    def __init__(self, model=None):
        self._caches = {}
        self.assertions = []
        self._model = model
        if model:
            self._model = model
        elif SelfExporter.default_model:
            self._model = SelfExporter.default_model
        else:
            raise Exception('A model must be specified explicitly if the '
                            'PySB self-exporter is not in use')

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, model):
        """ Changing the model invalidates any caches """
        self._caches = {}
        self._model = model

    def _ensure_required_caches(self, assertion):
        for cache in assertion.required_caches:
            if cache not in self._KNOWN_CACHES.keys():
                raise Exception('Unknown assertion cache: %s' % cache)
            self._caches[cache] = self._KNOWN_CACHES[cache](self.model)

    def add(self, assertion):
        self.assertions.append(assertion)

    def check(self, assertion):
        """
        Checks an assertion immediately without adding it to the test suite

        Parameters
        ----------
        assertion: ModelAssertion
            An instance of the ModelAssertion subclass

        Returns
        -------
        True if assertion succeeded or raises a ModelAssertionFailure
        exception if not
        """

        self._ensure_required_caches(assertion)
        return assertion.check(self.model, **{name: self._caches[name]
                                              for name in
                                              assertion.required_caches})

    def check_all(self, stop_on_exception=False):
        """Runs all assertions in the test suite"""
        for a in self.assertions:
            print('%s... ' % repr(a), end="")
            try:
                self.check(a)
                print('%sOK%s' % (self._COL['OK'], self._COL['END']))
            except ModelAssertionFailure as e:
                print('%sFAIL%s' % (self._COL['FAIL'], self._COL['END']))
                print('  ' + str(e.message))
                if stop_on_exception:
                    return
            except Exception as e:
                print('%sERROR%s' % (self._COL['FAIL'], self._COL['END']))
                print('  ' + str(e))
                if stop_on_exception:
                    return


class ModelAssertion(object):
    """
    Base class for model assertions
    """
    @abc.abstractmethod
    def __init__(self, *args, **kwargs):
        self.required_caches = set()
        self._last_result = None

    def __repr__(self):
        return '%s()' % self.__class__.__name__

    @abc.abstractmethod
    def check(self, model, **kwargs):
        if not isinstance(model, pysb.Model):
            raise ValueError('model should be an instance of pysb.Model')


def _negated_subclass(assertion_class, class_name):
    """
    Creates a negated version of an assertion class through subclassing

    Parameters
    ----------
    assertion_class: class
        ModelAssertion or its subclasses (not an instance)

    Returns
    -------
    A negated version of the ModelAssertion class

    """
    class NegatedSubclass(assertion_class):
        """
        Negated version of :class:`.{}`
        """.format(assertion_class.__class__.__name__)

        def check(self, model, **kwargs):
            try:
                super(self.__class__, self).check(model, **kwargs)
            except ModelAssertionFailure:
                return True

            raise ModelAssertionFailure(assertion=self, model=model,
                                        message=self._last_result)

    NegatedSubclass.__name__ = class_name
    return NegatedSubclass


class ReactionNetworkAssertion(ModelAssertion):
    """
    Base class for reaction network assertions

    Checks the reaction network has been generated
    """
    def __init__(self, *args, **kwargs):
        super(ReactionNetworkAssertion, self).__init__(*args, **kwargs)

    @abc.abstractmethod
    def check(self, model, **kwargs):
        super(ReactionNetworkAssertion, self).check(model)
        if not model.species:
            raise ModelAssertionFailure(assertion=self,
                                        model=model,
                                        message='Reaction network has not '
                                                'been generated yet')


class SpeciesAssertion(ReactionNetworkAssertion):
    """
    Class for checking species within a reaction network
    """
    def __init__(self, *args, **kwargs):
        super(SpeciesAssertion, self).__init__(*args, **kwargs)
        self.required_caches.add('species_pattern_matcher')
        self.pattern = args[0]

    @abc.abstractmethod
    def check(self, model, **kwargs):
        super(SpeciesAssertion, self).check(model)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.pattern)


class SpeciesExists(SpeciesAssertion):
    """
    Checks a species pattern exists in the list of species
    """
    def check(self, model, **kwargs):
        self._last_result = kwargs['species_pattern_matcher'].match(
            self.pattern)
        if self._last_result:
            return True
        else:
            raise ModelAssertionFailure(assertion=self, model=model)


SpeciesDoesNotExist = _negated_subclass(SpeciesExists, 'SpeciesDoesNotExist')


class ReactionAssertion(ReactionNetworkAssertion):
    def __init__(self, *args, **kwargs):
        super(ReactionAssertion, self).__init__(*args, **kwargs)
        self.required_caches.add('reaction_pattern_matcher')
        self.pattern = args[0]

    @abc.abstractmethod
    def check(self, model, **kwargs):
        super(ReactionAssertion, self).check(model)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.pattern)


class SpeciesIsProduct(ReactionAssertion):
    """
    Checks a species pattern appears on the product side of a reaction
    """
    def check(self, model, **kwargs):
        self._last_result = kwargs[
            'reaction_pattern_matcher'].match_products(self.pattern)
        if not self._last_result:
            raise ModelAssertionFailure(assertion=self, model=model)
        else:
            return True

SpeciesNeverProduct = _negated_subclass(SpeciesIsProduct, 'SpeciesNeverProduct')


class SpeciesIsReactant(ReactionAssertion):
    """
    Checks a species pattern appears on the reactant side of a reaction
    """
    def check(self, model, **kwargs):
        self._last_result = kwargs[
            'reaction_pattern_matcher'].match_reactants(self.pattern)
        if self._last_result:
            raise ModelAssertionFailure(assertion=self, model=model,
                                        message=self._last_result)
        else:
            return True

SpeciesNeverReactant = _negated_subclass(SpeciesIsReactant,
                                         'SpeciesNeverReactant')


class SpeciesOnlyProduct(ReactionAssertion):
    """ Checks a species appears as a product but never as a reactant """
    def check(self, model, **kwargs):
        rpm = kwargs['reaction_pattern_matcher']
        if not rpm.match_products(self.pattern):
            raise ModelAssertionFailure(assertion=self,
                                        model=model,
                                        message='Does not appear as product')
        as_reactant = rpm.match_reactants(self.pattern)
        if as_reactant:
            raise ModelAssertionFailure(assertion=self,
                                        model=model,
                                        message='Appears as reactant:' +
                                                str(as_reactant))

        return True


class SpeciesOnlyReactant(ReactionAssertion):
    """ Checks a species appears as a reactant but never as a product """

    def check(self, model, **kwargs):
        rpm = kwargs['reaction_pattern_matcher']
        if not rpm.match_reactants(self.pattern):
            raise ModelAssertionFailure(assertion=self,
                                        model=model,
                                        message='Does not appear as reactant')
        as_product = rpm.match_products(self.pattern)
        if as_product:
            raise ModelAssertionFailure(assertion=self,
                                        model=model,
                                        message='Appears as product: ' +
                                                str(as_product))

        return True


class RuleAssertion(ModelAssertion):
    def __init__(self, *args, **kwargs):
        super(RuleAssertion, self).__init__(*args, **kwargs)
        self.required_caches.add('rule_pattern_matcher')
        self.pattern = args[0]

    @abc.abstractmethod
    def check(self, model, **kwargs):
        super(RuleAssertion, self).check(model)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self.pattern)


class AllObservablesInRules(RuleAssertion):
    def check(self, model, **kwargs):
        rpm = kwargs['rule_pattern_matcher']
        unmatched_observables = []
        for obs in model.observables:
            matches = rpm.match_rules(obs.reaction_pattern)
            if not matches:
                unmatched_observables.append(obs)
        if unmatched_observables:
            raise ModelAssertionFailure(assertion=self,
                                        model=model,
                                        message='Unmatched Observables: ' +
                                                str(unmatched_observables))

        return True



from pysb.testing import *
from pysb import *
from pysb.core import as_complex_pattern
from pysb.bng import *
import os
import unittest
from nose.tools import assert_raises_regexp
from pysb.generator.bng import BngPrinter
import sympy
import math


@with_model
def test_generate_network():
    Monomer('A')
    assert_raises((NoInitialConditionsError, NoRulesError),
                  generate_network, model)
    Parameter('A_0', 1)
    Initial(A(), A_0)
    assert_raises(NoRulesError, generate_network, model)
    Parameter('k', 1)
    Rule('degrade', A() >> None, k)
    ok_(generate_network(model))


@unittest.skipIf(os.name == 'nt', 'BNG Console does not work on Windows')
@with_model
def test_simulate_network_console():
    Monomer('A')
    Parameter('A_0', 1)
    Initial(A(), A_0)
    Parameter('k', 1)
    Rule('degrade', A() >> None, k)
    with BngConsole(model) as bng:
        bng.generate_network()
        bng.action('simulate', method='ssa', t_end=20000, n_steps=100, netfile='pysb.net')


@with_model
def test_sequential_simulations():
    Monomer('A')
    Parameter('A_0', 1)
    Initial(A(), A_0)
    Parameter('k', 1)
    Rule('degrade', A() >> None, k)
    # Suppress network overwrite warning from simulate command
    with BngFileInterface(model) as bng:
        bng.action('generate_network')
        bng.action('simulate', method='ssa', t_end=20000, n_steps=100)
        bng.execute()
        yfull1 = bng.read_simulation_results()
        ok_(yfull1.size == 101)

        # Run another simulation by reloading the existing network file
        bng.action('simulate', method='ssa', t_end=10000, n_steps=50)
        bng.execute(reload_netfile=True)
        yfull2 = bng.read_simulation_results()
        ok_(yfull2.size == 51)


@with_model
def test_compartment_species_equivalence():
    Parameter('p', 1)
    Monomer('Q', ['x'])
    Monomer('R', ['y'])
    Compartment('C', None, 3, p)
    Rule('bind', Q(x=None) + R(y=None) >> Q(x=1) % R(y=1), p)
    Initial(Q(x=None) ** C, p)
    Initial(R(y=None) ** C, p)
    generate_equations(model)
    for i, ic in enumerate(model.initials):
        ok_(ic.pattern.is_equivalent_to(model.species[i]))
    ok_(model.species[2].is_equivalent_to(Q(x=1) ** C % R(y=1) ** C))


@with_model
def test_bidirectional_rules_collapse():
    Monomer('A')
    Monomer('B')
    Initial(B(), Parameter('B_init', 0))
    Initial(A(), Parameter('A_init', 100))
    Rule('Rule1', B() | A(), Parameter('k3', 10), Parameter('k1', 1))
    Rule('Rule2', A() | B(), k1, Parameter('k2', 1))
    Rule('Rule3', B() >> A(), Parameter('k4', 5))
    generate_equations(model)
    ok_(len(model.reactions) == 4)
    ok_(len(model.reactions_bidirectional) == 1)
    ok_(len(model.reactions_bidirectional[0]['rule']) == 3)
    ok_(model.reactions_bidirectional[0]['reversible'])


@with_model
def test_bidirectional_rules():
    Monomer('A')
    Monomer('B')
    Initial(A(), Parameter('A_init', 100))
    Rule('Rule1', A() | B(), Parameter('k1', 1), Parameter('k2', 1))
    Rule('Rule2', B() >> A(), Parameter('k3', 10))
    Rule('Rule3', B() >> A(), Parameter('k4', 5))
    generate_equations(model)
    ok_(len(model.reactions) == 4)
    ok_(len(model.reactions_bidirectional) == 1)
    ok_(len(model.reactions_bidirectional[0]['rule']) == 3)
    ok_(model.reactions_bidirectional[0]['reversible'])
    #TODO Check that 'rate' has 4 terms


@with_model
def test_zero_order_synth_no_initials():
    Monomer('A')
    Monomer('B')
    Rule('Rule1', None >> A(), Parameter('ksynth', 100))
    Rule('Rule2', A() | B(), Parameter('kf', 10), Parameter('kr', 1))
    generate_equations(model)


@with_model
def test_reversible_synth_deg():
    Monomer('A')
    Parameter('k_synth', 2.0)
    Parameter('k_deg', 1.0)
    Rule('synth_deg', A() | None, k_deg, k_synth)
    assert synth_deg.is_synth()
    assert synth_deg.is_deg()
    generate_equations(model)


@with_model
def test_nfsim():
    Monomer('A', ['a'])
    Monomer('B', ['b'])

    Parameter('ksynthA', 100)
    Parameter('ksynthB', 100)
    Parameter('kbindAB', 100)

    Parameter('A_init', 20)
    Parameter('B_init', 30)

    Initial(A(a=None), A_init)
    Initial(B(b=None), B_init)

    Observable("A_free", A(a=None))
    Observable("B_free", B(b=None))
    Observable("AB_complex", A(a=1) % B(b=1))

    Rule('A_synth', None >> A(a=None), ksynthA)
    Rule('B_synth', None >> B(b=None), ksynthB)
    Rule('AB_bind', A(a=None) + B(b=None) >> A(a=1) % B(b=1), kbindAB)

    with BngFileInterface(model) as bng:
        bng.action('simulate', method='nf', t_end=1000, n_steps=100)
        bng.execute()
        res = bng.read_simulation_results()
        assert res.dtype.names == ('time', 'A_free', 'B_free', 'AB_complex')
        assert len(res) == 101


@with_model
def test_unicode_strs():
    Monomer(u'A', [u'b'], {u'b':[u'y', u'n']})
    Monomer(u'B')
    Rule(u'rule1', A(b=u'y') >> B(), Parameter(u'k', 1))
    Initial(A(b=u'y'), Parameter(u'A_0', 100))
    generate_equations(model)


@with_model
def test_none_in_rxn_pat():
    Monomer(u'A', [u'b'], {u'b': [u'y', u'n']})
    Monomer(u'B')
    Rule(u'rule1', A(b=u'y') + None >> None + B(),
         Parameter(u'k', 1))
    Initial(A(b=u'y'), Parameter(u'A_0', 100))
    generate_equations(model)


@with_model
def test_sympy_parameter_keyword():
    Monomer('A')
    Initial(A(), Parameter('A_0', 100))
    Parameter('deg', 10)  # deg is a sympy function
    Rule('Rule1', A() >> None, deg)
    generate_equations(model)


@with_model
def test_log():
    Monomer('A')
    Expression('expr1', sympy.log(100))
    Rule('Rule1', None >> A(), expr1)
    generate_equations(model)


@with_model
def test_bng_error():
    Monomer('A', ['a'], {'a': ['s1', 's2']})
    Parameter('A_0', 100)
    Initial(A(a='s1'), A_0)
    Parameter('kf', 1)
    # The following rule does not specify A's site on the RHS, so should generate a BNG error
    Rule('r1', A(a='s1') >> A(), kf)
    assert_raises_regexp(
        BngInterfaceError,
        'Molecule created in reaction rule: Component\(s\) a missing from molecule A\(\)',
        generate_equations,
        model
    )


@with_model
def test_fixed_species():
    Monomer('A', ['a'])
    Monomer('B', ['b'])
    Initial(A(a=1) % B(b=1), Parameter('AB_0', 1), fixed=True)
    Rule('rule1', A(a=1) % B(b=1) >> A(a=None) + B(b=None), Parameter('k', 1))
    generate_equations(model)
    num_non_zeros = model.stoichiometry_matrix[0].getnnz()
    assert num_non_zeros == 0


@with_model
def test_multistate():
    Monomer('A', ['a', 'a'], {'a': ['u', 'p']})
    Parameter('k1', 100)
    Parameter('A_0', 200)
    Rule('r1', None >> A(a=MultiState('u', 'p')), k1)
    Initial(A(a=MultiState(('u', 1), 'p')) %
            A(a=MultiState(('u', 1), 'u')), A_0)

    generate_equations(model)

    assert model.species[0].is_equivalent_to(
        A(a=MultiState(('u', 1), 'p')) % A(a=MultiState(('u', 1), 'u')))
    assert model.species[1].is_equivalent_to(
        as_complex_pattern(A(a=MultiState('u', 'p'))))


@with_model
def test_multibonds():
    Monomer('A', ['a'])
    Monomer('B', ['b'])
    Parameter('k1', 100)
    Parameter('A_0', 200)
    Parameter('B_0', 50)
    Rule('r1', A(a=None) + A(a=None) + B(b=None) >>
            A(a=1) % A(a=[1, 2]) % B(b=2), k1)
    Initial(A(a=None), A_0)
    Initial(B(b=None), B_0)

    generate_equations(model)

    assert model.species[2].is_equivalent_to(
        A(a=1) % A(a=[1, 2]) % B(b=2)
    )


def _bng_print(expr):
    return BngPrinter(order='none').doprint(expr)


def _rint(num):
    """ This is how BNG/muparser does "rounding" """
    return math.floor(num + 0.5)


def test_bng_rounding():
    """ Test Ceiling and Floor match BNG implementation of rounding """
    x_sym = sympy.symbols('x')
    for x in (-1.5, -1, -0.5, 0, 0.5, 1.0, 1.5):
        for expr in (sympy.ceiling(x_sym), sympy.floor(x_sym)):
            assert expr.subs({'x': x}) == eval(
                _bng_print(expr), {}, {'rint': _rint, 'x': x})


def test_bng_printer():
    # Constants
    assert _bng_print(sympy.pi) == '_pi'
    assert _bng_print(sympy.E) == '_e'

    x, y = sympy.symbols('x y')

    # Binary functions
    assert _bng_print(sympy.sympify('x & y')) == 'x && y'
    assert _bng_print(sympy.sympify('x | y')) == 'x || y'

    # Trig functions
    assert _bng_print(sympy.sin(x)) == 'sin(x)'
    assert _bng_print(sympy.cos(x)) == 'cos(x)'
    assert _bng_print(sympy.tan(x)) == 'tan(x)'
    assert _bng_print(sympy.asin(x)) == 'asin(x)'
    assert _bng_print(sympy.acos(x)) == 'acos(x)'
    assert _bng_print(sympy.atan(x)) == 'atan(x)'
    assert _bng_print(sympy.sinh(x)) == 'sinh(x)'
    assert _bng_print(sympy.cosh(x)) == 'cosh(x)'
    assert _bng_print(sympy.tanh(x)) == 'tanh(x)'
    assert _bng_print(sympy.asinh(x)) == 'asinh(x)'
    assert _bng_print(sympy.acosh(x)) == 'acosh(x)'
    assert _bng_print(sympy.atanh(x)) == 'atanh(x)'

    # Logs and powers
    assert _bng_print(sympy.log(x)) == 'ln(x)'
    assert _bng_print(sympy.exp(x)) == 'exp(x)'
    assert _bng_print(sympy.sqrt(x)) == 'sqrt(x)'

    # Rounding
    assert _bng_print(sympy.Abs(x)) == 'abs(x)'
    assert _bng_print(sympy.floor(x)) == 'rint(x - 0.5)'
    assert _bng_print(sympy.ceiling(x)) == '(rint(x + 1) - 1)'

    # Min/max
    assert _bng_print(sympy.Min(x, y)) == 'min(x, y)'
    assert _bng_print(sympy.Max(x, y)) == 'max(x, y)'

    # Relational
    assert _bng_print(sympy.Eq(x, y)) == 'x == y'
    assert _bng_print(sympy.Ne(x, y)) == 'x != y'
    assert _bng_print(x < y) == 'x < y'
    assert _bng_print(x <= y) == 'x <= y'
    assert _bng_print(x > y) == 'x > y'
    assert _bng_print(x >= y) == 'x >= y'


def test_bng_printer_relational_unknown():
    class NewRelational(sympy.core.relational.Relational):
        rel_op = "??????????"  # A highly unlikely rel_op for a new subclass.
    x = sympy.Symbol("x")
    assert_raises(NotImplementedError, _bng_print, NewRelational(x, x))


def test_parse_bngl_expression_if():
    x, y = sympy.symbols('x y')
    assert parse_bngl_expr('if(x>y, 1, 3)') == \
        sympy.Piecewise((1, x > y), (3, True))


def test_parse_bngl_expression_exponentiate():
    x, y = sympy.symbols('x y')
    assert parse_bngl_expr('x ^ y') == sympy.Pow(x, y)


def test_parse_bngl_expression_and_or_equals():
    x, y = sympy.symbols('x y')
    assert parse_bngl_expr('x and y') == sympy.And(x, y)
    assert parse_bngl_expr('x or y') == sympy.Or(x, y)
    assert parse_bngl_expr('x == y') == sympy.Eq(x, y)


def test_bng_boolean_multiply_number():
    assert parse_bngl_expr('(2 > 1) * 4') == 4
    assert parse_bngl_expr('4 * (2 > 1)') == 4



"""
A class for converting a PySB model to a set of ordinary differential
equations for integration in MATLAB.

Note that for use in MATLAB, the name of the ``.m`` file must match the name of
the exported MATLAB class (e.g., ``robertson.m`` for the example below).

For information on how to use the model exporters, see the documentation
for :py:mod:`pysb.export`.

Output for the Robertson example model
======================================

Information on the form and usage of the generated MATLAB class is contained in
the documentation for the MATLAB model, as shown in the following example for
``pysb.examples.robertson``::

    classdef robertson
        % A simple three-species chemical kinetics system known as "Robertson's
        % example", as presented in:
        % 
        % H. H. Robertson, The solution of a set of reaction rate equations, in Numerical
        % Analysis: An Introduction, J. Walsh, ed., Academic Press, 1966, pp. 178-182.
        % 
        % A class implementing the ordinary differential equations
        % for the robertson model.
        %
        % Save as robertson.m.
        %
        % Generated by pysb.export.matlab.MatlabExporter.
        %
        % Properties
        % ----------
        % observables : struct
        %     A struct containing the names of the observables from the
        %     PySB model as field names. Each field in the struct
        %     maps the observable name to a matrix with two rows:
        %     the first row specifies the indices of the species
        %     associated with the observable, and the second row
        %     specifies the coefficients associated with the species.
        %     For any given timecourse of model species resulting from
        %     integration, the timecourse for an observable can be
        %     retrieved using the get_observable method, described
        %     below.
        %
        % parameters : struct
        %     A struct containing the names of the parameters from the
        %     PySB model as field names. The nominal values are set by
        %     the constructor and their values can be overriden
        %     explicitly once an instance has been created.
        %
        % Methods
        % -------
        % robertson.odes(tspan, y0)
        %     The right-hand side function for the ODEs of the model,
        %     for use with MATLAB ODE solvers (see Examples).
        %
        % robertson.get_initial_values()
        %     Returns a vector of initial values for all species,
        %     specified in the order that they occur in the original
        %     PySB model (i.e., in the order found in model.species).
        %     Non-zero initial conditions are specified using the
        %     named parameters included as properties of the instance.
        %     Hence initial conditions other than the defaults can be
        %     used by assigning a value to the named parameter and then
        %     calling this method. The vector returned by the method
        %     is used for integration by passing it to the MATLAB
        %     solver as the y0 argument.
        %
        % robertson.get_observables(y)
        %     Given a matrix of timecourses for all model species
        %     (i.e., resulting from an integration of the model),
        %     get the trajectories corresponding to the observables.
        %     Timecourses are returned as a struct which can be
        %     indexed by observable name.
        %
        % Examples
        % --------
        % Example integration using default initial and parameter
        % values:
        %
        % >> m = robertson();
        % >> tspan = [0 100];
        % >> [t y] = ode15s(@m.odes, tspan, m.get_initial_values());
        %
        % Retrieving the observables:
        %
        % >> y_obs = m.get_observables(y)
        %
        properties
            observables
            parameters
        end

        methods
            function self = robertson()
                % Assign default parameter values
                self.parameters = struct( ...
                    'k1', 0.040000000000000001, ...
                    'k2', 30000000, ...
                    'k3', 10000, ...
                    'A_0', 1, ...
                    'B_0', 0, ...
                    'C_0', 0);

                % Define species indices (first row) and coefficients
                % (second row) of named observables
                self.observables = struct( ...
                    'A_total', [1; 1], ...
                    'B_total', [2; 1], ...
                    'C_total', [3; 1]);
            end

            function initial_values = get_initial_values(self)
                % Return the vector of initial conditions for all
                % species based on the values of the parameters
                % as currently defined in the instance.

                initial_values = zeros(1,3);
                initial_values(1) = self.parameters.A_0; % A()
                initial_values(2) = self.parameters.B_0; % B()
                initial_values(3) = self.parameters.C_0; % C()
            end

            function y = odes(self, tspan, y0)
                % Right hand side function for the ODEs

                % Shorthand for the struct of model parameters
                p = self.parameters;

                % A();
                y(1,1) = -p.k1*y0(1) + p.k3*y0(2)*y0(3);
                % B();
                y(2,1) = p.k1*y0(1) - p.k2*power(y0(2), 2) - p.k3*y0(2)*y0(3);
                % C();
                y(3,1) = p.k2*power(y0(2), 2);
            end

            function y_obs = get_observables(self, y)
                % Retrieve the trajectories for the model observables
                % from a matrix of the trajectories of all model
                % species.

                % Initialize the struct of observable timecourses
                % that we will return
                y_obs = struct();

                % Iterate over the observables;
                observable_names = fieldnames(self.observables);
                for i = 1:numel(observable_names)
                    obs_matrix = self.observables.(observable_names{i});
                    if isempty(obs_matrix)
                        y_obs.(observable_names{i}) = zeros(size(y, 1), 1);
                        continue
                    end
                    species = obs_matrix(1, :);
                    coefficients = obs_matrix(2, :);
                    y_obs.(observable_names{i}) = ...
                                    y(:, species) * coefficients';
                end
            end
        end
    end
"""

import pysb
import pysb.bng
import sympy
import re
from io import StringIO
from pysb.export import Exporter, pad, ExpressionsNotSupported, \
    CompartmentsNotSupported


class MatlabExporter(Exporter):
    """A class for returning the ODEs for a given PySB model for use in
    MATLAB.

    Inherits from :py:class:`pysb.export.Exporter`, which implements
    basic functionality for all exporters.
    """
    def export(self):
        """Generate a MATLAB class definition containing the ODEs for the PySB
        model associated with the exporter.

        Returns
        -------
        string
            String containing the MATLAB code for an implementation of the
            model's ODEs.
        """
        if self.model.expressions:
            raise ExpressionsNotSupported()
        if self.model.compartments:
            raise CompartmentsNotSupported()

        output = StringIO()
        pysb.bng.generate_equations(self.model)

        docstring = ''
        if self.docstring:
            docstring += self.docstring.replace('\n', '\n    % ')

        # Substitute underscores for any dots in the model name
        model_name = self.model.name.replace('.', '_')

        # -- Parameters and Initial conditions -------
        # Declare the list of parameters as a struct
        params_str = 'self.parameters = struct( ...\n'+' '*16
        params_str_list = []
        for i, p in enumerate(self.model.parameters):
            # Add parameter to struct along with nominal value
            cur_p_str = "'%s', %.17g" % (_fix_underscores(p.name), p.value)
            # Decide whether to continue or terminate the struct declaration:
            if i == len(self.model.parameters) - 1:
                cur_p_str += ');'    # terminate
            else:
                cur_p_str += ', ...' # continue

            params_str_list.append(cur_p_str)

        # Format and indent the params struct declaration
        params_str += ('\n'+' '*16).join(params_str_list)

        # Fill in an array of the initial conditions based on the named
        # parameter values
        initial_values_str = ('initial_values = zeros(1,%d);\n'+' '*12) % \
                             len(self.model.species)
        initial_values_str += ('\n'+' '*12).join(
                ['initial_values(%d) = self.parameters.%s; %% %s' %
                 (i+1, _fix_underscores(ic.value.name), ic.pattern)
                 for i, ic in enumerate(self.model.initials)])

        # -- Build observables declaration --
        observables_str = 'self.observables = struct( ...\n'+' '*16
        observables_str_list = []
        for i, obs in enumerate(self.model.observables):
            # Associate species and coefficient lists with observable names,
            # changing from zero- to one-based indexing
            cur_obs_str = "'%s', [%s; %s]" % \
                          (_fix_underscores(obs.name),
                           ' '.join([str(sp+1) for sp in obs.species]),
                           ' '.join([str(c) for c in obs.coefficients]))
            # Decide whether to continue or terminate the struct declaration:
            if i == len(self.model.observables) - 1:
                cur_obs_str += ');'    # terminate
            else:
                cur_obs_str += ', ...' # continue

            observables_str_list.append(cur_obs_str)
        # Format and indent the observables struct declaration
        observables_str += ('\n'+' '*16).join(observables_str_list)

        # -- Build ODEs -------
        # Build a stringified list of species
        species_list = ['%% %s;' % s for i, s in enumerate(self.model.species)]
        # Build the ODEs as strings from the model.odes array
        odes_list = ['y(%d,1) = %s;' % (i+1, sympy.ccode(self.model.odes[i])) 
                     for i in range(len(self.model.odes))] 
        # Zip the ODEs and species string lists and then flatten them
        # (results in the interleaving of the two lists)
        odes_species_list = [item for sublist in zip(species_list, odes_list)
                                  for item in sublist]
        # Flatten to a string and add correct indentation
        odes_str = ('\n'+' '*12).join(odes_species_list)

        # Change species names from, e.g., '__s(0)' to 'y0(1)' (note change
        # from zero-based indexing to 1-based indexing)
        odes_str = re.sub(r'__s(\d+)', \
                          lambda m: 'y0(%s)' % (int(m.group(1))+1), odes_str)
        # Change C code 'pow' function to MATLAB 'power' function
        odes_str = re.sub(r'pow\(', 'power(', odes_str)
        # Prepend 'p.' to named parameters and fix any underscores
        for i, p in enumerate(self.model.parameters):
            odes_str = re.sub(r'\b(%s)\b' % p.name,
                              'p.%s' % _fix_underscores(p.name), odes_str)

        # -- Build final output --
        output.write(pad(r"""
            classdef %(model_name)s
                %% %(docstring)s
                %% A class implementing the ordinary differential equations
                %% for the %(model_name)s model.
                %%
                %% Save as %(model_name)s.m.
                %%
                %% Generated by pysb.export.matlab.MatlabExporter.
                %%
                %% Properties
                %% ----------
                %% observables : struct
                %%     A struct containing the names of the observables from the
                %%     PySB model as field names. Each field in the struct
                %%     maps the observable name to a matrix with two rows:
                %%     the first row specifies the indices of the species
                %%     associated with the observable, and the second row
                %%     specifies the coefficients associated with the species.
                %%     For any given timecourse of model species resulting from
                %%     integration, the timecourse for an observable can be
                %%     retrieved using the get_observable method, described
                %%     below.
                %%
                %% parameters : struct
                %%     A struct containing the names of the parameters from the
                %%     PySB model as field names. The nominal values are set by
                %%     the constructor and their values can be overriden
                %%     explicitly once an instance has been created.
                %%
                %% Methods
                %% -------
                %% %(model_name)s.odes(tspan, y0)
                %%     The right-hand side function for the ODEs of the model,
                %%     for use with MATLAB ODE solvers (see Examples).
                %%
                %% %(model_name)s.get_initial_values()
                %%     Returns a vector of initial values for all species,
                %%     specified in the order that they occur in the original
                %%     PySB model (i.e., in the order found in model.species).
                %%     Non-zero initial conditions are specified using the
                %%     named parameters included as properties of the instance.
                %%     Hence initial conditions other than the defaults can be
                %%     used by assigning a value to the named parameter and then
                %%     calling this method. The vector returned by the method
                %%     is used for integration by passing it to the MATLAB
                %%     solver as the y0 argument.
                %%
                %% %(model_name)s.get_observables(y)
                %%     Given a matrix of timecourses for all model species
                %%     (i.e., resulting from an integration of the model),
                %%     get the trajectories corresponding to the observables.
                %%     Timecourses are returned as a struct which can be
                %%     indexed by observable name.
                %%
                %% Examples
                %% --------
                %% Example integration using default initial and parameter
                %% values:
                %%
                %% >> m = %(model_name)s();
                %% >> tspan = [0 100];
                %% >> [t y] = ode15s(@m.odes, tspan, m.get_initial_values());
                %%
                %% Retrieving the observables:
                %%
                %% >> y_obs = m.get_observables(y)
                %%
                properties
                    observables
                    parameters
                end

                methods
                    function self = %(model_name)s()
                        %% Assign default parameter values
                        %(params_str)s

                        %% Define species indices (first row) and coefficients
                        %% (second row) of named observables
                        %(observables_str)s
                    end

                    function initial_values = get_initial_values(self)
                        %% Return the vector of initial conditions for all
                        %% species based on the values of the parameters
                        %% as currently defined in the instance.

                        %(initial_values_str)s
                    end

                    function y = odes(self, tspan, y0)
                        %% Right hand side function for the ODEs

                        %% Shorthand for the struct of model parameters
                        p = self.parameters;

                        %(odes_str)s
                    end

                    function y_obs = get_observables(self, y)
                        %% Retrieve the trajectories for the model observables
                        %% from a matrix of the trajectories of all model
                        %% species.

                        %% Initialize the struct of observable timecourses
                        %% that we will return
                        y_obs = struct();

                        %% Iterate over the observables;
                        observable_names = fieldnames(self.observables);
                        for i = 1:numel(observable_names)
                            obs_matrix = self.observables.(observable_names{i});
                            if isempty(obs_matrix)
                                y_obs.(observable_names{i}) = zeros(size(y, 1), 1);
                                continue
                            end
                            species = obs_matrix(1, :);
                            coefficients = obs_matrix(2, :);
                            y_obs.(observable_names{i}) = ...
                                            y(:, species) * coefficients';
                        end
                    end
                end
            end
            """, 0) %
            {'docstring': docstring,
             'model_name': model_name,
             'params_str':params_str,
             'initial_values_str': initial_values_str,
             'observables_str': observables_str,
             'params_str': params_str,
             'odes_str': odes_str})

        return output.getvalue()

def _fix_underscores(name):
    if name.startswith('_'):
        return 'X' + name
    else:
        return name



"""Adapted from a portion of the model published in:

Input-output behavior of ErbB signaling pathways as revealed by a mass action
model trained against dynamic data. William W Chen, Birgit Schoeberl, Paul J
Jasper, Mario Niepel, Ulrik B Nielsen, Douglas A Lauffenburger & Peter K
Sorger. Mol Syst Biol. 2009;5:239. Epub 2009 Jan 20. doi:10.1038/msb.2008.74

http://www.nature.com/msb/journal/v5/n1/full/msb200874.html

Implemented by: Jeremy Muhlich
"""

from __future__ import print_function
from pysb import *
from pysb.macros import catalyze_state

Model()

Monomer('Ras', ['k'])
Annotation(Ras, 'http://identifiers.org/uniprot/P01116', 'hasPart')
Annotation(Ras, 'http://identifiers.org/uniprot/P01112', 'hasPart')
Annotation(Ras, 'http://identifiers.org/uniprot/P01111', 'hasPart')
Monomer('Raf', ['s', 'k'], {'s': ['u', 'p']})
Annotation(Raf, 'http://identifiers.org/uniprot/P15056', 'hasPart')
Annotation(Raf, 'http://identifiers.org/uniprot/P04049', 'hasPart')
Annotation(Raf, 'http://identifiers.org/uniprot/P10398', 'hasPart')
Monomer('MEK', ['s218', 's222', 'k'], {'s218': ['u', 'p'], 's222': ['u', 'p']})
Annotation(MEK, 'http://identifiers.org/uniprot/Q02750', 'hasPart')
Annotation(MEK, 'http://identifiers.org/uniprot/P36507', 'hasPart')
Monomer('ERK', ['t185', 'y187'], {'t185': ['u', 'p'], 'y187': ['u', 'p']})
Annotation(ERK, 'http://identifiers.org/uniprot/P27361', 'hasPart')
Annotation(ERK, 'http://identifiers.org/uniprot/P28482', 'hasPart')
Monomer('PP2A', ['ppt'])
Annotation(PP2A, 'http://identifiers.org/mesh/24544')
Monomer('MKP', ['ppt'])
Annotation(MKP, 'http://identifiers.org/mesh/24536')

# Use generic rates for forward/reverse binding and kinase/phosphatase catalysis
kf_bind = 1e-5
kr_bind = 1e-1
kcat_phos = 1e-1
kcat_dephos = 3e-3

# Build handy rate "sets"
klist_bind = [kf_bind, kr_bind]
klist_phos = klist_bind + [kcat_phos]
klist_dephos = klist_bind + [kcat_dephos]

def mapk_single(kinase, pptase, substrate, site):
    """Kinase phos/dephosphorylation."""
    ppt_substrate = substrate()
    if 'k' in ppt_substrate.monomer.sites:
        # Ensure substrates which are themselves kinases don't get
        # dephosphorylated while they are bound to *their* substrate.
        ppt_substrate = ppt_substrate(k=None)
    components = catalyze_state(kinase, 'k',
                                substrate, site, site, 'u', 'p',
                                klist_phos)
    components |= catalyze_state(pptase, 'ppt',
                                 ppt_substrate, site, site, 'p', 'u',
                                 klist_dephos)
    return components

def mapk_double(kinase, pptase, substrate, site1, site2):
    """Distributive + ordered double kinase phos/dephosphorylation."""
    components = mapk_single(kinase, pptase, substrate({site2: 'u'}), site1)
    components |= mapk_single(kinase, pptase, substrate({site1: 'p'}), site2)
    return components

# Ras-Raf-MEK-ERK kinase cascade
mapk_single(Ras, PP2A, Raf, 's')
mapk_double(Raf(s='p'), PP2A, MEK, 's218', 's222')
mapk_double(MEK(s218='p', s222='p'), MKP, ERK, 't185', 'y187')

Initial(Ras(k=None), Parameter('Ras_0', 6e4))
Initial(Raf(s='u', k=None), Parameter('Raf_0', 7e4))
Initial(MEK(s218='u', s222='u', k=None), Parameter('MEK_0', 3e6))
Initial(ERK(t185='u', y187='u'), Parameter('ERK_0', 7e5))
Initial(PP2A(ppt=None), Parameter('PP2A_0', 2e5))
Initial(MKP(ppt=None), Parameter('MKP_0', 1.7e4))

Observable('ppMEK', MEK(s218='p', s222='p'))
Observable('ppERK', ERK(t185='p', y187='p'))


if __name__ == '__main__':
    print(__doc__, "\n", model)
    print("""
NOTE: This model code is designed to be imported and programatically
manipulated, not executed directly. The above output is merely a
diagnostic aid.""")



from pysb.pattern import SpeciesPatternMatcher, match_complex_pattern
from pysb.examples import robertson, bax_pore, bax_pore_sequential, \
    earm_1_3, kinase_cascade, bngwiki_egfr_simple
from pysb.bng import generate_equations
from nose.tools import assert_raises
from pysb import as_complex_pattern, as_reaction_pattern, ANY, WILD, \
    Monomer
import collections


def test_species_pattern_matcher():
    # See also SpeciesPatternMatcher doctests

    # Check that SpeciesPatternMatcher raises exception if model has no species
    model = robertson.model
    model.reset_equations()
    assert_raises(Exception, SpeciesPatternMatcher, model)

    model = bax_pore.model
    generate_equations(model)
    spm = SpeciesPatternMatcher(model)
    BAX = model.monomers['BAX']
    sp_sets = spm.species_fired_by_reactant_pattern(
        as_reaction_pattern(BAX(t1=None, t2=None))
    )
    assert len(sp_sets) == 1
    assert len(sp_sets[0]) == 2

    sp_sets = spm.species_fired_by_reactant_pattern(
        as_reaction_pattern(BAX(t1=WILD, t2=ANY))
    )
    assert len(sp_sets) == 1
    assert len(sp_sets[0]) == 10


def test_wildcards():
    a_wild = as_complex_pattern(Monomer('A', ['b'], _export=False)(b=WILD))
    b_mon = as_complex_pattern(Monomer('B', _export=None))

    # B() should not match A(b=WILD)
    assert not match_complex_pattern(b_mon, a_wild)


def check_all_species_generated(model):
    """Check the lists of rules and triggering species match BNG"""
    generate_equations(model)
    spm = SpeciesPatternMatcher(model)

    # Get the rules and species firing them using the pattern matcher
    species_produced = dict()
    for rule, rp_species in spm.rule_firing_species().items():
        species_produced[rule.name] = set()
        for sp_list in rp_species:
            for sp in sp_list:
                species_produced[rule.name].add(
                    model.get_species_index(sp))

    # Get the equivalent dictionary of {rule name: [reactant species]} from BNG
    species_bng = collections.defaultdict(set)
    for rxn in model.reactions:
        for rule in rxn['rule']:
            species_bng[rule].update(rxn['reactants'])

    # Our output should match BNG
    assert species_produced == species_bng


def test_all_species_generated():
    for model in [bax_pore, earm_1_3, bax_pore_sequential, kinase_cascade,
                  bngwiki_egfr_simple]:
        yield (check_all_species_generated, model.model)



from pysb.tools.sensitivity_analysis import \
    PairwiseSensitivity, InitialsSensitivity
from pysb.examples.tyson_oscillator import model
import numpy as np
import numpy.testing as npt
import os
from pysb.simulator.scipyode import ScipyOdeSimulator
import tempfile
import shutil
from nose.tools import raises


class TestSensitivityAnalysis(object):
    def setUp(self):
        self.tspan = np.linspace(0, 200, 5001)
        self.observable = 'Y3'
        self.savename = 'sens_out_test'
        self.output_dir = tempfile.mkdtemp()
        self.vals = np.linspace(.8, 1.2, 5)
        self.vals = [.8, .9, 1., 1.1, 1.2]
        self.model = model
        self.solver = ScipyOdeSimulator(self.model,
                                        tspan=self.tspan,
                                        integrator='lsoda',
                                        integrator_options={'rtol': 1e-8,
                                                            'atol': 1e-8,
                                                            'mxstep': 20000})
        self.sens = PairwiseSensitivity(
            solver=self.solver,
            values_to_sample=self.vals,
            objective_function=self.obj_func_cell_cycle,
            observable=self.observable
        )

        self.p_simulated = np.array(
            [[0., 0., 0., 0., 0., 5.0301, 2.6027, 0., -2.5118, -4.5758],
             [0., 0., 0., 0., 0., 5.0301, 2.5832, 0., -2.5313, -4.5823],
             [0., 0., 0., 0., 0., 5.0301, 2.5767, 0., -2.5313, -4.5953],
             [0., 0., 0., 0., 0., 5.0301, 2.5767, 0., -2.5313, -4.6082],
             [0., 0., 0., 0., 0., 5.0301, 2.5767, 0., -2.5313, -4.60829],
             [5.0301, 5.0301, 5.0301, 5.0301, 5.0301, 0., 0., 0., 0., 0.],
             [2.6027, 2.5832, 2.5767, 2.5767, 2.5767, 0., 0., 0., 0., 0.],
             [0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
             [-2.5118, -2.5313, -2.5313, -2.5313, -2.5313, 0., 0., 0., 0., 0.],
             [-4.5758, -4.5823, -4.5953, -4.6082, -4.6082, 0., 0., 0., 0.,
              0.]])

        self.sens.run()

    def tearDown(self):
        shutil.rmtree(self.output_dir)

    def obj_func_cell_cycle(self, out):
        timestep = self.tspan[:-1]
        y = out[:-1] - out[1:]
        freq = 0
        local_times = []
        prev = y[0]
        for n in range(1, len(y)):
            if y[n] > 0 > prev:
                local_times.append(timestep[n])
                freq += 1
            prev = y[n]

        local_times = np.array(local_times)
        local_freq = np.average(local_times) / len(local_times) * 2
        return local_freq

    def test_run(self):

        sens_vode = PairwiseSensitivity(
            solver=self.solver,
            values_to_sample=self.vals,
            objective_function=self.obj_func_cell_cycle,
            observable=self.observable
        )
        sens_vode.run()
        npt.assert_almost_equal(self.sens.p_matrix, self.p_simulated,
                                decimal=3)

    def test_old_class_naming(self):
        sens_vode = InitialsSensitivity(
            solver=self.solver,
            values_to_sample=self.vals,
            objective_function=self.obj_func_cell_cycle,
            observable=self.observable
        )
        sens_vode.run()
        npt.assert_almost_equal(self.sens.p_matrix, self.p_simulated,
                                decimal=3)

    def test_p_matrix_shape(self):
        assert self.sens.p_matrix.shape == (10, 10)

    def test_p_matrix(self):
        npt.assert_almost_equal(self.sens.p_matrix, self.p_simulated,
                                decimal=3)

    def test_pmatrix_outfile_exists(self):
        self.sens.run(save_name=self.savename,
                      out_dir=self.output_dir)
        assert os.path.exists(os.path.join(
            self.output_dir, '{}_p_matrix.csv'.format(self.savename)
        ))
        assert os.path.exists(os.path.join(
            self.output_dir, '{}_p_prime_matrix.csv'.format(self.savename)
        ))

    def test_create_png(self):
        self.sens.create_boxplot_and_heatplot(save_name='test',
                                              out_dir=self.output_dir)

        assert os.path.exists(os.path.join(self.output_dir, 'test.png'))
        assert os.path.exists(os.path.join(self.output_dir, 'test.eps'))
        assert os.path.exists(os.path.join(self.output_dir, 'test.svg'))

        self.sens.create_individual_pairwise_plots(save_name='test2',
                                                   out_dir=self.output_dir)
        assert os.path.exists(os.path.join(self.output_dir,
                                           'test2_subplots.png'))

        self.sens.create_plot_p_h_pprime(save_name='test3',
                                         out_dir=self.output_dir)
        assert os.path.exists(os.path.join(self.output_dir,
                                           'test3_P_H_P_prime.png'))

    def test_unique_simulations_only(self):
        vals = [.8, .9, 1.1, 1.2, 1.3]
        sens = PairwiseSensitivity(
            values_to_sample=vals,
            objective_function=self.obj_func_cell_cycle,
            observable=self.observable,
            solver=self.solver
        )
        sens.run()
        self.sens.create_plot_p_h_pprime(save_name='test4',
                                         out_dir=self.output_dir)
        assert os.path.exists(os.path.join(self.output_dir,
                                           'test4_P_H_P_prime.png'))

    def test_param_pair(self):
        vals = [.9, 1.0, 1.1]
        sens = PairwiseSensitivity(
            values_to_sample=vals,
            objective_function=self.obj_func_cell_cycle,
            observable=self.observable,
            solver=self.solver,
            sample_list=['k1', 'cdc0']
        )
        sens.run()
        self.sens.create_plot_p_h_pprime(save_name='test4',
                                         out_dir=self.output_dir)
        assert os.path.exists(os.path.join(self.output_dir,
                                           'test4_P_H_P_prime.png'))

    def test_all_params(self):
        vals = [.9, 1.1]

        sens = PairwiseSensitivity(
            values_to_sample=vals,
            objective_function=self.obj_func_cell_cycle,
            observable=self.observable,
            solver=self.solver,
            sens_type='all'
        )
        sens.run()
        self.sens.create_plot_p_h_pprime(save_name='test4',
                                         out_dir=None)
        assert os.path.exists('test4_P_H_P_prime.png')

    @raises(ValueError)
    def test_param_not_in_model(self):
        vals = [.8, .9, 1.1, 1.2, 1.3]
        solver = ScipyOdeSimulator(self.model,
                                   tspan=self.tspan,
                                   integrator='lsoda',
                                   integrator_options={'rtol': 1e-8,
                                                       'atol': 1e-8,
                                                       'mxstep': 20000})
        sens = PairwiseSensitivity(
            values_to_sample=vals,
            objective_function=self.obj_func_cell_cycle,
            observable=self.observable,
            solver=solver, sample_list=['a0']
        )

    @raises(ValueError)
    def test_sens_type_and_list_none(self):
        vals = [.8, .9, 1.1, 1.2, 1.3]
        sens = PairwiseSensitivity(
            values_to_sample=vals,
            objective_function=self.obj_func_cell_cycle,
            observable=self.observable,
            solver=self.solver, sample_list=None, sens_type=None
        )

    @raises(TypeError)
    def test_bad_solver(self):
        vals = [.8, .9, 1.1, 1.2, 1.3]
        sens = PairwiseSensitivity(
            values_to_sample=vals,
            objective_function=self.obj_func_cell_cycle,
            observable=self.observable,
            solver=None, sample_list=None, sens_type=None
        )



from pysb.testing import *
import numpy as np
from pysb import Monomer, Parameter, Initial, Observable, Rule, Expression
from pysb.simulator.bng import BngSimulator, PopulationMap
from pysb.bng import generate_equations
from pysb.examples import robertson, expression_observables

_BNG_SEED = 123


class TestBngSimulator(object):
    @with_model
    def setUp(self):
        Monomer('A', ['a'])
        Monomer('B', ['b'])

        Parameter('ksynthA', 100)
        Parameter('ksynthB', 100)
        Parameter('kbindAB', 100)

        Parameter('A_init', 0)
        Parameter('B_init', 0)

        Initial(A(a=None), A_init)
        Initial(B(b=None), B_init)

        Observable("A_free", A(a=None))
        Observable("B_free", B(b=None))
        Observable("AB_complex", A(a=1) % B(b=1))

        Rule('A_synth', None >> A(a=None), ksynthA)
        Rule('B_synth', None >> B(b=None), ksynthB)
        Rule('AB_bind', A(a=None) + B(b=None) >> A(a=1) % B(b=1), kbindAB)

        self.model = model
        generate_equations(self.model)

        # Convenience shortcut for accessing model monomer objects
        self.mon = lambda m: self.model.monomers[m]

        # This timespan is chosen to be enough to trigger a Jacobian evaluation
        # on the various solvers.
        self.time = np.linspace(0, 1)
        self.sim = BngSimulator(self.model, tspan=self.time)

    def test_1_simulation(self):
        x = self.sim.run()
        assert np.allclose(x.tout, self.time)

    def test_multi_simulations(self):
        x = self.sim.run(n_runs=10)
        assert np.shape(x.observables) == (10, 50)
        # Check initials are getting correctly reset on each simulation
        assert np.allclose(x.species[-1][0, :], x.species[0][0, :])

    def test_change_parameters(self):
        x = self.sim.run(n_runs=10, param_values={'ksynthA': 200},
                         initials={self.model.species[0]: 100})
        species = np.array(x.all)
        assert species[0][0][0] == 100.

    def test_bng_pla(self):
        self.sim.run(n_runs=5, method='pla', seed=_BNG_SEED)

    def test_tout_matches_tspan(self):
        # Linearly spaced, starting from 0
        assert all(self.sim.run(tspan=[0, 10, 20]).tout[0] == [0, 10, 20])
        # Non-linearly spaced, starting from 0
        assert all(self.sim.run(tspan=[0, 10, 30]).tout[0] == [0, 10, 30])
        # Linearly spaced, starting higher than 0
        assert all(self.sim.run(tspan=[10, 20, 30]).tout[0] == [10, 20, 30])
        # Linearly spaced, starting higher than 0
        assert all(self.sim.run(tspan=[5, 20, 30]).tout[0] == [5, 20, 30])

    def tearDown(self):
        self.model = None
        self.time = None
        self.sim = None


def test_bng_ode_with_expressions():
    model = expression_observables.model
    model.reset_equations()

    sim = BngSimulator(model, tspan=np.linspace(0, 1))
    x = sim.run(n_runs=1, method='ode')
    assert len(x.expressions) == 50
    assert len(x.observables) == 50


class TestNfSim(object):
    def setUp(self):
        self.model = robertson.model
        self.model.reset_equations()
        self.sim = BngSimulator(self.model, tspan=np.linspace(0, 1))
        self.param_values_2sets = [[1, 2, 3, 4, 5, 6], [7, 8, 9, 10, 11, 12]]

    def test_nfsim_2runs(self):
        x = self.sim.run(n_runs=1, method='nf', seed=_BNG_SEED)
        observables = np.array(x.observables)
        assert len(observables) == 50

        A = self.model.monomers['A']
        x = self.sim.run(n_runs=2, method='nf', tspan=np.linspace(0, 1),
                    initials={A(): 100}, seed=_BNG_SEED)
        assert (x.nsims == 2)
        assert np.allclose(x.dataframe.loc[0, 0.0], [100.0, 0.0, 0.0])

    def test_nfsim_2initials(self):
        # Test with two initials
        A = self.model.monomers['A']
        x2 = self.sim.run(method='nf', tspan=np.linspace(0, 1),
                     initials={A(): [100, 200]}, seed=_BNG_SEED)
        assert x2.nsims == 2
        assert np.allclose(x2.dataframe.loc[0, 0.0], [100.0, 0.0, 0.0])
        assert np.allclose(x2.dataframe.loc[1, 0.0], [200.0, 0.0, 0.0])

    def test_nfsim_2params(self):
        # Test with two param_values
        x3 = self.sim.run(method='nf', tspan=np.linspace(0, 1),
                          param_values=self.param_values_2sets, seed=_BNG_SEED)
        assert x3.nsims == 2
        assert np.allclose(x3.param_values, self.param_values_2sets)

    def test_nfsim_2initials_2params(self):
        # Test with two initials and two param_values
        A = self.model.monomers['A']
        x = self.sim.run(method='nf',
                         tspan=np.linspace(0, 1),
                         initials={A(): [101, 201]},
                         param_values=[[1, 2, 3, 4, 5, 6],
                                       [7, 8, 9, 10, 11, 12]],
                         seed=_BNG_SEED)
        assert x.nsims == 2
        # Initials for A should be set by initials dict, and from param_values
        # for B and C
        assert np.allclose(x.dataframe.loc[0, 0.0], [101.0, 5.0, 6.0])
        assert np.allclose(x.dataframe.loc[1, 0.0], [201.0, 11.0, 12.0])

    @raises(ValueError)
    def test_nfsim_different_initials_lengths(self):
        A = self.model.monomers['A']
        B = self.model.monomers['B']

        sim = BngSimulator(self.model, tspan=np.linspace(0, 1),
                           initials={B(): [150, 250, 350]})
        sim.run(initials={A(): [275, 375]})

    @raises(ValueError)
    def test_nfsim_different_initials_params_lengths(self):
        A = self.model.monomers['A']

        sim = BngSimulator(self.model, tspan=np.linspace(0, 1),
                           initials={A(): [150, 250, 350]},
                           param_values=self.param_values_2sets)


def test_hpp():
    model = robertson.model
    # Reset equations from any previous network generation
    model.reset_equations()

    A = robertson.model.monomers['A']
    klump = Parameter('klump', 10000, _export=False)
    model.add_component(klump)

    population_maps = [
        PopulationMap(A(), klump)
    ]

    sim = BngSimulator(model, tspan=np.linspace(0, 1))
    x = sim.run(n_runs=1, method='nf', population_maps=population_maps,
                seed=_BNG_SEED)
    observables = np.array(x.observables)
    assert len(observables) == 50


def test_stop_if():
    Model()
    Monomer('A')
    Rule('A_synth', None >> A(), Parameter('k', 1))
    Observable('Atot', A())
    Expression('exp_const', k + 1)
    Expression('exp_dyn', Atot + 1)
    sim = BngSimulator(model, verbose=5)
    tspan = np.linspace(0, 100, 101)
    x = sim.run(tspan, stop_if='Atot>9', seed=_BNG_SEED)
    # All except the last Atot value should be <=9
    assert all(x.observables['Atot'][:-1] <= 9)
    assert x.observables['Atot'][-1] > 9
    # Starting with Atot > 9 should terminate simulation immediately
    y = sim.run(tspan, initials=x.species[-1], stop_if='Atot>9')
    assert len(y.observables) == 1


def test_set_initials_by_params():
    # This tests setting initials by changing their underlying parameter values
    # BNG Simulator uses a dictionary for initials, unlike e.g.
    # ScipyOdeSimulator, so a separate test is needed

    model = robertson.model
    t = np.linspace(0, 40, 51)
    ic_params = model.parameters_initial_conditions()
    param_values = np.array([p.value for p in model.parameters])
    ic_mask = np.array([p in ic_params for p in model.parameters])

    bng_sim = BngSimulator(model, tspan=t, verbose=0)

    # set all initial conditions to 1
    param_values[ic_mask] = np.array([1, 1, 1])
    traj = bng_sim.run(param_values=param_values)

    # set properly here
    assert np.allclose(traj.initials, [1, 1, 1])

    # overwritten in bng file. lines 196-202.
    # Values from initials_dict are used, but it should take them from
    # self.initials, so I don't see how they are getting overwritten?
    print(traj.dataframe.loc[0])
    assert np.allclose(traj.dataframe.loc[0][0:3], [1, 1, 1])

    # Same here
    param_values[ic_mask] = np.array([0, 1, 1])
    traj = bng_sim.run(param_values=param_values)
    assert np.allclose(traj.initials, [0, 1, 1])
    assert np.allclose(traj.dataframe.loc[0][0:3], [0, 1, 1])



from pysb.testing import *
import sys
import copy
import numpy as np
from pysb import Monomer, Parameter, Initial, Observable, Rule, Expression
from pysb.simulator import ScipyOdeSimulator, InconsistentParameterError
from pysb.simulator.scipyode import CythonRhsBuilder
from pysb.examples import robertson, earm_1_0, tyson_oscillator, localfunc
import unittest
import pandas as pd


class TestScipySimulatorBase(object):
    @with_model
    def setUp(self):
        Monomer('A', ['a'])
        Monomer('B', ['b'])

        Parameter('ksynthA', 100)
        Parameter('ksynthB', 100)
        Parameter('kbindAB', 100)

        Parameter('A_init', 0)
        Parameter('B_init', 0)

        Initial(A(a=None), A_init)
        Initial(B(b=None), B_init)

        Observable("A_free", A(a=None))
        Observable("B_free", B(b=None))
        Observable("AB_complex", A(a=1) % B(b=1))

        Rule('A_synth', None >> A(a=None), ksynthA)
        Rule('B_synth', None >> B(b=None), ksynthB)
        Rule('AB_bind', A(a=None) + B(b=None) >> A(a=1) % B(b=1), kbindAB)

        self.model = model

        # Convenience shortcut for accessing model monomer objects
        self.mon = lambda m: self.model.monomers[m]

        # This timespan is chosen to be enough to trigger a Jacobian evaluation
        # on the various solvers.
        self.time = np.linspace(0, 1)
        self.sim = ScipyOdeSimulator(self.model, tspan=self.time,
                                     integrator='vode')

    def tearDown(self):
        self.model = None
        self.time = None
        self.sim = None


class TestScipySimulatorSingle(TestScipySimulatorBase):
    def test_vode_solver_run(self):
        """Test vode."""
        simres = self.sim.run()
        assert simres._nsims == 1

    @raises(ValueError)
    def test_invalid_init_kwarg(self):
        ScipyOdeSimulator(self.model, tspan=self.time, spam='eggs')

    def test_lsoda_solver_run(self):
        """Test lsoda."""
        solver_lsoda = ScipyOdeSimulator(self.model, tspan=self.time,
                                         integrator='lsoda')
        solver_lsoda.run()

    def test_lsoda_jac_solver_run(self):
        """Test lsoda and analytic jacobian."""
        solver_lsoda_jac = ScipyOdeSimulator(self.model, tspan=self.time,
                                             integrator='lsoda',
                                             use_analytic_jacobian=True)
        solver_lsoda_jac.run()

    def test_y0_as_list(self):
        """Test y0 with list of initial conditions"""
        # Test the initials getter method before anything is changed
        assert np.allclose(
            self.sim.initials[0][0:2],
            [ic.value.value for ic in self.model.initials]
        )

        initials = [10, 20, 0]
        simres = self.sim.run(initials=initials)
        assert np.allclose(simres.initials[0], initials)
        assert np.allclose(simres.observables['A_free'][0], 10)

    def test_y0_as_ndarray(self):
        """Test y0 with numpy ndarray of initial conditions"""
        simres = self.sim.run(initials=np.asarray([10, 20, 0]))
        assert np.allclose(simres.observables['A_free'][0], 10)

    def test_y0_as_dictionary_monomer_species(self):
        """Test y0 with model-defined species."""
        self.sim.initials = {self.mon('A')(a=None): 17}
        base_initials = self.sim.initials
        assert base_initials[0][0] == 17

        simres = self.sim.run(initials={self.mon('A')(a=None): 10,
                               self.mon('B')(b=1) % self.mon('A')(a=1): 0,
                               self.mon('B')(b=None): 0})
        assert np.allclose(simres.initials, [10, 0, 0])
        assert np.allclose(simres.observables['A_free'][0], 10)

        # Initials should reset to base values
        assert np.allclose(self.sim.initials, base_initials)

    def test_y0_as_dictionary_with_bound_species(self):
        """Test y0 with dynamically generated species."""
        simres = self.sim.run(initials={self.mon('A')(a=None): 0,
                               self.mon('B')(b=1) % self.mon('A')(a=1): 100,
                               self.mon('B')(b=None): 0})
        assert np.allclose(simres.observables['AB_complex'][0], 100)

    def test_y0_as_dataframe(self):
        initials_dict = {self.mon('A')(a=None): [0],
                         self.mon('B')(b=1) % self.mon('A')(a=1): [100],
                         self.mon('B')(b=None): [0]}
        initials_df = pd.DataFrame(initials_dict)
        simres = self.sim.run(initials=initials_df)
        assert np.allclose(simres.observables['AB_complex'][0], 100)

    @raises(ValueError)
    def test_y0_as_pandas_series(self):
        self.sim.run(initials=pd.Series())

    @raises(TypeError)
    def test_y0_non_numeric_value(self):
        """Test y0 with non-numeric value."""
        self.sim.run(initials={self.mon('A')(a=None): 'eggs'})

    def test_param_values_as_dictionary(self):
        """Test param_values as a dictionary."""
        simres = self.sim.run(param_values={'kbindAB': 0})
        # kbindAB=0 should ensure no AB_complex is produced.
        assert np.allclose(simres.observables["AB_complex"], 0)

    def test_param_values_as_dataframe(self):
        simres = self.sim.run(param_values=pd.DataFrame({'kbindAB': [0]}))
        assert np.allclose(simres.observables['AB_complex'], 0)

    @raises(ValueError)
    def test_param_values_as_pandas_series(self):
        self.sim.run(param_values=pd.Series())

    def test_param_values_as_list_ndarray(self):
        """Test param_values as a list and ndarray."""
        orig_param_values = self.sim.param_values
        param_values = [50, 60, 70, 0, 0]
        self.sim.param_values = param_values
        simres = self.sim.run()
        assert np.allclose(self.sim.param_values, param_values)
        assert np.allclose(simres.param_values, param_values)
        # Reset to original param values
        self.sim.param_values = orig_param_values

        # Same thing, but with a numpy array, applied as a run argument
        param_values = np.asarray([55, 65, 75, 0, 0])
        simres = self.sim.run(param_values=param_values)
        assert np.allclose(simres.param_values, param_values)
        # param_values should reset to originals after the run
        assert np.allclose(self.sim.param_values, orig_param_values)

    @raises(IndexError)
    def test_param_values_invalid_dictionary_key(self):
        """Test param_values with invalid parameter name."""
        self.sim.run(param_values={'spam': 150})

    @raises(ValueError, TypeError)
    def test_param_values_non_numeric_value(self):
        """Test param_values with non-numeric value."""
        self.sim.run(param_values={'ksynthA': 'eggs'})

    def test_result_dataframe(self):
        df = self.sim.run().dataframe


class TestScipyOdeCompilerTests(TestScipySimulatorBase):
    """Test vode and analytic jacobian with different compiler backends"""
    def setUp(self):
        super(TestScipyOdeCompilerTests, self).setUp()
        self.args = {'model': self.model,
                     'tspan': self.time,
                     'integrator': 'vode',
                     'use_analytic_jacobian': True}

        self.python_sim = ScipyOdeSimulator(compiler='python', **self.args)
        self.python_res = self.python_sim.run()

    def test_cython(self):
        sim = ScipyOdeSimulator(compiler='cython', **self.args)
        simres = sim.run()
        assert simres.species.shape[0] == self.args['tspan'].shape[0]
        assert np.allclose(self.python_res.dataframe, simres.dataframe)


class TestScipySimulatorSequential(TestScipySimulatorBase):
    def test_sequential_initials(self):
        simres = self.sim.run()
        orig_initials = self.sim.initials

        new_initials = [10, 20, 30]
        simres = self.sim.run(initials=new_initials)

        # Check that single-run initials applied properly to the result
        assert np.allclose(simres.species[0], new_initials)
        assert np.allclose(simres.initials, new_initials)
        # Check that the single-run initials were removed after the run
        assert np.allclose(self.sim.initials, orig_initials)

    def test_sequential_initials_dict_then_list(self):
        A, B = self.model.monomers

        base_sim = ScipyOdeSimulator(
            self.model,
            initials={A(a=None): 10, B(b=None): 20})

        assert np.allclose(base_sim.initials, [10, 20, 0])
        assert len(base_sim.initials_dict) == 2

        # Now set initials using a list, which should overwrite the dict
        base_sim.initials = [30, 40, 50]

        assert np.allclose(base_sim.initials, [30, 40, 50])
        assert np.allclose(
            sorted([x[0] for x in base_sim.initials_dict.values()]),
            base_sim.initials)

    def test_sequential_param_values(self):
        orig_param_values = self.sim.param_values
        new_param_values = {'kbindAB': 0}
        new_initials = [15, 25, 35]
        simres = self.sim.run(param_values=new_param_values,
                              initials=new_initials)
        # No new AB_complex should be formed
        assert np.allclose(simres.observables['AB_complex'], new_initials[2])
        assert simres.nsims == 1
        # Original param_values should be restored after run
        assert np.allclose(self.sim.param_values, orig_param_values)

        # Check that per-run param override works when a base param
        # dictionary is also specified
        self.sim.param_values = new_param_values
        base_param_values = new_param_values
        new_param_values = {'ksynthB': 50}
        simres = self.sim.run(param_values=new_param_values)
        # Check that new param value override applied
        assert np.allclose(simres.param_values[0][1],
                           new_param_values['ksynthB'])
        # Check that simulator reverts to base param values
        assert np.allclose(self.sim.param_values[0][2],
                           base_param_values['kbindAB'])
        # Reset to original param values
        self.sim.param_values = orig_param_values

    def test_sequential_tspan(self):
        tspan = np.linspace(0, 10, 11)
        orig_tspan = self.sim.tspan
        simres = self.sim.run(tspan=tspan)
        # Check that new tspan applied properly
        assert np.allclose(simres.tout, tspan)
        # Check that simulator instance reset to original tspan
        assert np.allclose(self.sim.tspan, orig_tspan)


class TestScipySimulatorMultiple(TestScipySimulatorBase):
    def test_initials_and_param_values_two_lists(self):
        initials = [[10, 20, 30], [50, 60, 70]]
        param_values = [[55, 65, 75, 0, 0],
                        [90, 100, 110, 5, 6]]
        import pysb.bng
        pysb.bng.generate_equations(self.sim.model)
        simres = self.sim.run(initials=initials, param_values=param_values)
        assert np.allclose(simres.species[0][0], initials[0])
        assert np.allclose(simres.species[1][0], initials[1])

        assert np.allclose(simres.param_values[0], param_values[0])
        assert np.allclose(simres.param_values[1], param_values[1])

        assert simres.nsims == 2

        # Check the methods underlying these properties work
        df = simres.dataframe
        all = simres.all

        # Try overriding above lists of initials/params with dicts
        self.sim.initials = initials
        self.sim.param_values = param_values
        simres = self.sim.run(
            initials={self.mon('A')(a=None): [103, 104]},
            param_values={'ksynthA': [101, 102]})
        # Simulator initials and params should not persist run() overrides
        assert np.allclose(self.sim.initials, initials)
        assert np.allclose(self.sim.param_values, param_values)
        # Create the expected initials/params arrays and compare to result
        initials = np.array(initials)
        initials[:, 0] = [103, 104]
        param_values = np.array(param_values)
        param_values[:, 0] = [101, 102]
        assert np.allclose(simres.initials, initials)
        assert np.allclose(simres.param_values, param_values)

    @raises(ValueError)
    def test_run_initials_different_length_to_base(self):
        initials = [[10, 20, 30, 40], [50, 60, 70, 80]]
        self.sim.initials = initials
        self.sim.run(initials=initials[0])

    @raises(ValueError)
    def test_run_params_different_length_to_base(self):
        param_values = [[55, 65, 75, 0, 0, 1],
                        [90, 100, 110, 5, 6, 7]]
        self.sim.param_values = param_values
        self.sim.run(param_values=param_values[0])

    @raises(InconsistentParameterError)
    def test_run_params_inconsistent_parameter_list(self):
        param_values = [55, 65, 75, 0, -3]
        self.sim.param_values = param_values
        self.sim.run(param_values=param_values[0])

    @raises(InconsistentParameterError)
    def test_run_params_inconsistent_parameter_dict(self):
        param_values = {'A_init': [0, -4]}
        self.sim.param_values = param_values
        self.sim.run(param_values=param_values[0])

    def test_param_values_dict(self):
        param_values = {'A_init': [0, 100]}
        initials = {self.model.monomers['B'](b=None): [250, 350]}

        simres = self.sim.run(param_values=param_values)
        assert np.allclose(simres.dataframe.loc[(slice(None), 0.0), 'A_free'],
                           [0, 100])

        simres = self.sim.run(param_values={'B_init': [200, 300]})
        assert np.allclose(simres.dataframe.loc[(slice(None), 0.0), 'A_free'],
                           [0, 0])
        assert np.allclose(simres.dataframe.loc[(slice(None), 0.0), 'B_free'],
                           [200, 300])

        simres = self.sim.run(initials=initials, param_values=param_values)
        assert np.allclose(simres.dataframe.loc[(slice(None), 0.0), 'A_free'],
                           [0, 100])
        assert np.allclose(simres.dataframe.loc[(slice(None), 0.0), 'B_free'],
                           [250, 350])

    @raises(ValueError)
    def test_initials_and_param_values_differing_lengths(self):
        initials = [[10, 20, 30, 40], [50, 60, 70, 80]]
        param_values = [[55, 65, 75, 0, 0],
                        [90, 100, 110, 5, 6],
                        [90, 100, 110, 5, 6]]
        self.sim.run(initials=initials, param_values=param_values)

    @unittest.skipIf(sys.version_info.major < 3,
                     'Parallel execution requires Python >= 3.3')
    def test_parallel(self):
        for integrator in ('vode', 'lsoda'):
            for use_analytic_jacobian in (True, False):
                yield self._check_parallel, integrator, use_analytic_jacobian

    def _check_parallel(self, integrator, use_analytic_jacobian):
        initials = [[10, 20, 30], [50, 60, 70]]
        sim = ScipyOdeSimulator(
            self.model, self.sim.tspan,
            initials=initials,
            integrator=integrator,
            use_analytic_jacobian=use_analytic_jacobian
        )
        base_res = sim.run(initials=initials)
        res = sim.run(initials=initials, num_processors=2)
        assert np.allclose(res.species, base_res.species)


@with_model
def test_integrate_with_expression():
    """Ensure a model with Expressions simulates."""

    Monomer('s1')
    Monomer('s9')
    Monomer('s16')
    Monomer('s20')

    # Parameters should be able to contain s(\d+) without error
    Parameter('ks0',2e-5)
    Parameter('ka20', 1e5)

    Initial(s9(), Parameter('s9_0', 10000))

    Observable('s1_obs', s1())
    Observable('s9_obs', s9())
    Observable('s16_obs', s16())
    Observable('s20_obs', s20())

    Expression('keff', (ks0*ka20)/(ka20+s9_obs))

    Rule('R1', None >> s16(), ks0)
    Rule('R2', None >> s20(), ks0)
    Rule('R3', s16() + s20() >> s16() + s1(), keff)

    time = np.linspace(0, 40)
    sim = ScipyOdeSimulator(model, tspan=time)
    simres = sim.run()
    keff_vals = simres.expressions['keff']
    assert len(keff_vals) == len(time)
    assert np.allclose(keff_vals, 1.8181818181818182e-05)


def test_set_initial_to_zero():
    sim = ScipyOdeSimulator(robertson.model, tspan=np.linspace(0, 100))
    simres = sim.run(initials={robertson.model.monomers['A'](): 0})
    assert np.allclose(simres.observables['A_total'], 0)


def test_robertson_integration():
    """Ensure robertson model simulates."""
    t = np.linspace(0, 100)
    sim = ScipyOdeSimulator(robertson.model, tspan=t, compiler="python")
    simres = sim.run()
    assert simres.species.shape[0] == t.shape[0]
    # Also run with cython compiler if available.
    if CythonRhsBuilder.check_safe():
        sim = ScipyOdeSimulator(robertson.model, tspan=t, compiler="cython")
        simres = sim.run()
        assert simres.species.shape[0] == t.shape[0]


def test_earm_integration():
    """Ensure earm_1_0 model simulates."""
    t = np.linspace(0, 1e3)
    sim = ScipyOdeSimulator(earm_1_0.model, tspan=t, compiler="python")
    sim.run()
    # Also run with cython compiler if available.
    if CythonRhsBuilder.check_safe():
        ScipyOdeSimulator(earm_1_0.model, tspan=t, compiler="cython").run()


@raises(ValueError)
def test_simulation_no_tspan():
    ScipyOdeSimulator(robertson.model).run()


@raises(UserWarning)
def test_nonexistent_integrator():
    """Ensure nonexistent integrator raises."""
    ScipyOdeSimulator(robertson.model, tspan=np.linspace(0, 1, 2),
                      integrator='does_not_exist')


def test_unicode_obsname_ascii():
    """Ensure ascii-convetible unicode observable names are handled."""
    t = np.linspace(0, 100)
    rob_copy = copy.deepcopy(robertson.model)
    rob_copy.observables[0].name = u'A_total'
    sim = ScipyOdeSimulator(rob_copy)
    simres = sim.run(tspan=t)
    simres.all
    simres.dataframe


if sys.version_info[0] < 3:
    @raises(ValueError)
    def test_unicode_obsname_nonascii():
        """Ensure non-ascii unicode observable names error in python 2."""
        t = np.linspace(0, 100)
        rob_copy = copy.deepcopy(robertson.model)
        rob_copy.observables[0].name = u'A_total\u1234'
        sim = ScipyOdeSimulator(rob_copy)
        simres = sim.run(tspan=t)


def test_unicode_exprname_ascii():
    """Ensure ascii-convetible unicode expression names are handled."""
    t = np.linspace(0, 100)
    rob_copy = copy.deepcopy(robertson.model)
    ab = rob_copy.observables['A_total'] + rob_copy.observables['B_total']
    expr = Expression(u'A_plus_B', ab, _export=False)
    rob_copy.add_component(expr)
    sim = ScipyOdeSimulator(rob_copy)
    simres = sim.run(tspan=t)
    simres.all
    simres.dataframe


if sys.version_info[0] < 3:
    @raises(ValueError)
    def test_unicode_exprname_nonascii():
        """Ensure non-ascii unicode expression names error in python 2."""
        t = np.linspace(0, 100)
        rob_copy = copy.deepcopy(robertson.model)
        ab = rob_copy.observables['A_total'] + rob_copy.observables['B_total']
        expr = Expression(u'A_plus_B\u1234', ab, _export=False)
        rob_copy.add_component(expr)
        sim = ScipyOdeSimulator(rob_copy)
        simres = sim.run(tspan=t)


def test_multiprocessing_lambdify():
    model = tyson_oscillator.model
    pars = [p.value for p in model.parameters]
    tspan = np.linspace(0, 100, 100)
    ScipyOdeSimulator(
        model, tspan=tspan, compiler='python',
        use_analytic_jacobian=True
    ).run(param_values=[pars, pars], num_processors=2)


def test_lambdify_localfunc():
    model = localfunc.model
    ScipyOdeSimulator(model, tspan=range(100), compiler='python').run()



# import abc
# import copy
# from pysb.logging import get_logger
#
#
# class MergeStrategy(abc.ABC):
#     """ Base class for building model merging strategies """
#     def __init__(self, model1, model2):
#         self._log = get_logger(__name__, model=model1)
#         self.m1 = model1
#         self.m2 = model2
#         self.m1_mapping = None
#         self.m2_mapping = None
#
#     @abc.abstractmethod
#     def merge(self):
#         """ Merge model1 and model2 to return a new, combined model """
#         pass
#
#     def _rename(self, source_namespace, target_namespace, name):
#         """ Propose a new name for a component in target_namespace """
#         new_name = name
#         i = 1
#         while new_name in source_namespace or new_name in target_namespace:
#             new_name = '{}__{}'.format(name, i)
#             i += 1
#
#         return new_name
#
#     def _combine_models(self):
#         m_out = copy.deepcopy(self.m1)
#         m_out.name += '_merge_{}'.format(self.m2.name)
#         target_namespace = set(m_out.components.keys())
#         source_namespace = set(self.m2.components.keys())
#         m2_copy = copy.deepcopy(self.m2)
#
#         self.m1_mapping = dict(zip(source_namespace, source_namespace))
#         self.m2_mapping = {}
#
#         for c in m2_copy.components:
#             if c.name in target_namespace:
#                 c_name = self._rename(source_namespace,
#                                       target_namespace,
#                                       c.name)
#                 self._log.debug('{} {} was renamed to {}'.format(
#                     c.__class__.__name__, c.name, c_name
#                 ))
#                 c.rename(c_name)
#             else:
#                 c_name = c.name
#             target_namespace.add(c_name)
#             self.m2_mapping[c.name] = c_name
#             m_out.add_component(c)
#
#         # Copy initial conditions
#         m_out.initial_conditions += m2_copy.initial_conditions
#
#         # Copy annotations
#         m_out.annotations += m2_copy.annotations
#
#         return m_out
#
#
# class NoOverlapMerge(MergeStrategy):
#     """ Naive strategy in which all components are assumed to be distinct """
#
#     def merge(self):
#         return self._combine_models()



"""Example of how to write a model without using SelfExporter.
"""

from __future__ import print_function
from pysb import *
import pysb.core

# disable SelfExporter from doing its thing
pysb.core.SelfExporter.do_export = False

model = Model('explicit')

L = Monomer('L', ['r'])
R = Monomer('R', ['l'])
model.add_component(L)
model.add_component(R)

kf = Parameter('kf', 1e-6)
L_binds_R = Rule('L_binds_R', L(r=None) + R(l=None) >> L(r=1) % R(l=1), kf)
model.add_component(kf)
model.add_component(L_binds_R)

L_0 = Parameter('L_0', 1e2)
R_0 = Parameter('R_0', 1e4)
model.add_component(L_0)
model.add_component(R_0)
model.add_initial(Initial(L(r=None), L_0))
model.add_initial(Initial(R(l=None), R_0))

if __name__ == '__main__':
    print(__doc__, "\n", model)
    print("""
NOTE: This model code is designed to be imported and programatically
manipulated, not executed directly. The above output is merely a
diagnostic aid.""")



# # from pysb.examples.earm_1_3 import model as earm
# from pysb.examples.bax_pore import model as bax_pore
# from pysb.importers.bngl import model_from_bngl
# from pysb.tests.test_importers import _bngl_location
# from pysb.merge import NoOverlapMerge
# from pysb.export.pysb_flat import PysbFlatExporter
#
# def test_merge_basic():
#     # egfr = model_from_bngl(_bngl_location('egfr_net'))
#     # print(egfr.monomers)
#     # print(earm.monomers)
#     # print(bax_pore.monomers)
#     m_merge = NoOverlapMerge(bax_pore, bax_pore).merge()
#     print(PysbFlatExporter(m_merge).export())
#     raise




"""
Module containing a class for exporting a PySB model to JSON

For information on how to use the model exporters, see the documentation
for :py:mod:`pysb.export`.
"""

from pysb.export import Exporter
from pysb.bng import generate_equations
import json
from pysb.core import Model, MultiState, KeywordMeta, Parameter, Expression


class JsonExporter(Exporter):
    """A class for returning the JSON for a given PySB model.

    Inherits from :py:class:`pysb.export.Exporter`, which implements
    basic functionality for all exporters.
    """

    def export(self, include_netgen=False):
        """Generate the corresponding JSON for the PySB model associated
        with the exporter.

        Parameters
        ----------
        include_netgen: bool
            Include cached network generation data (reactions, species,
            local function-derived parameters and expressions) if True.

        Returns
        -------
        string
            The JSON output for the model.
        """
        return json.dumps(self.model, cls=PySBJSONWithNetworkEncoder
                          if include_netgen else PySBJSONEncoder)


class PySBJSONEncoder(json.JSONEncoder):
    """
    Encode a PySB model in JSON

    This encoder stores the model without caching the reaction network. To also
    store the reaction network, see :py:class:`PySBJSONWithNetworkEncoder`.

    Attributes correspond to their PySB equivalents (monomers, parameters, etc.)
    and are mostly stored verbatim, with the following exceptions.

    * MultiStates and the ANY and WILD state values use a special object format
    * References to other components are stored using the component name
    * Sympy expressions are encoded as strings using the default encoder

    The protocol number (currently: 1) specifies semantic model compatibility,
    and should be incremented if new features are added which affect how a model
    is simulated or prevent a model being loaded by
    :py:func:`pysb.importers.json.PySBJSONDecoder`.
    """
    PROTOCOL = 1

    @classmethod
    def encode_keyword(cls, keyword):
        return {'__object__': str(keyword)}

    @classmethod
    def encode_multistate(cls, stateval):
        return {
            '__object__': '__multistate__',
            'sites': stateval.sites
        }

    @classmethod
    def encode_monomer(cls, mon):
        return {
            'name': mon.name,
            'sites': mon.sites,
            'states': mon.site_states
        }

    @classmethod
    def encode_compartment(cls, cpt):
        return {
            'name': cpt.name,
            'parent': cpt.parent.name if cpt.parent else None,
            'dimension': cpt.dimension,
            'size': cpt.size.name if cpt.size else None
        }

    @classmethod
    def encode_parameter(cls, par):
        return {
            'name': par.name,
            'value': par.value
        }

    @classmethod
    def encode_expression(cls, expr):
        return {
            'name': expr.name,
            'expr': expr.expr.name if isinstance(
                expr.expr, (Parameter, Expression)) else str(expr.expr)
        }

    @classmethod
    def encode_monomer_pattern(cls, mp):
        return {
            'monomer': mp.monomer.name,
            'site_conditions': mp.site_conditions,
            'compartment': mp.compartment.name if mp.compartment else None,
            'tag': mp._tag.name if mp._tag else None
        }

    @classmethod
    def encode_complex_pattern(cls, cp):
        return {
            'monomer_patterns': [cls.encode_monomer_pattern(mp)
                                 for mp in cp.monomer_patterns],
            'compartment': cp.name if cp.compartment else None,
            'match_once': cp.match_once,
            'tag': cp._tag.name if cp._tag else None
        }

    @classmethod
    def encode_reaction_pattern(cls, rp):
        return {
            'complex_patterns': [cls.encode_complex_pattern(cp)
                                 for cp in rp.complex_patterns]
        }

    @classmethod
    def encode_rule_expression(cls, rexp):
        return {
            'reactant_pattern': cls.encode_reaction_pattern(
                rexp.reactant_pattern),
            'product_pattern': cls.encode_reaction_pattern(
                rexp.product_pattern),
            'reversible': rexp.is_reversible
        }

    @classmethod
    def encode_rule(cls, r):
        return {
            'name': r.name,
            'rule_expression': cls.encode_rule_expression(r.rule_expression),
            'rate_forward': r.rate_forward.name,
            'rate_reverse': r.rate_reverse.name if r.rate_reverse else None,
            'delete_molecules': r.delete_molecules,
            'move_connected': r.move_connected
        }

    @classmethod
    def encode_observable(cls, obs):
        return {
            'name': obs.name,
            'reaction_pattern': cls.encode_reaction_pattern(
                obs.reaction_pattern),
            'match': obs.match
        }

    @classmethod
    def encode_initial(cls, init):
        return {
            'pattern': cls.encode_complex_pattern(init.pattern),
            'parameter_or_expression': init.value.name,
            'fixed': init.fixed
        }

    @classmethod
    def encode_annotation(cls, ann):
        return {
            'subject': 'model' if isinstance(ann.subject, Model)
            else ann.subject.name,
            'object': str(ann.object),
            'predicate': str(ann.predicate)
        }

    @classmethod
    def encode_tag(cls, tag):
        return {
            'name': tag.name
        }

    @classmethod
    def encode_model(cls, model):
        d = dict(protocol=cls.PROTOCOL, name=model.name)

        encoders = {
            'monomers': cls.encode_monomer,
            'compartments': cls.encode_compartment,
            'tags': cls.encode_tag,
            'parameters': cls.encode_parameter,
            'expressions': cls.encode_expression,
            'rules': cls.encode_rule,
            'observables': cls.encode_observable,
            'initials': cls.encode_initial,
            'annotations': cls.encode_annotation
        }

        for component_type, encoder in encoders.items():
            d[component_type] = [encoder(component)
                                 for component in
                                 getattr(model, component_type)]

        return d

    def default(self, o):
        if isinstance(o, Model):
            return self.encode_model(o)
        elif isinstance(o, MultiState):
            return self.encode_multistate(o)
        elif isinstance(o, KeywordMeta):
            return self.encode_keyword(o)

        return super(PySBJSONEncoder, self).default(o)


class PySBJSONWithNetworkEncoder(PySBJSONEncoder):
    """
    Encode a PySB model and its reaction network in JSON

    This encoder stores the model including the cached reaction network. To
    encode the model without the reaction network, see
    :py:class:`PySBJSONEncoder`, which also includes implementation details.
    """

    @classmethod
    def encode_reaction(cls, rxn):
        rxn = rxn.copy()
        rxn['rate'] = rxn['rate'].name \
            if isinstance(rxn['rate'], (Parameter, Expression)) \
            else str(rxn['rate'])
        return rxn

    @classmethod
    def encode_observable(cls, obs):
        o = super(PySBJSONWithNetworkEncoder, cls).encode_observable(obs)
        o['species'] = obs.species
        o['coefficients'] = obs.coefficients
        return o

    @classmethod
    def encode_model(cls, model):
        d = super(PySBJSONWithNetworkEncoder, cls).encode_model(model)

        # Ensure network generation has taken place
        generate_equations(model)

        additional_encoders = {
            '_derived_parameters': cls.encode_parameter,
            '_derived_expressions': cls.encode_expression,
            'reactions': cls.encode_reaction,
            'reactions_bidirectional': cls.encode_reaction,
            'species': cls.encode_complex_pattern
        }

        for component_type, encoder in additional_encoders.items():
            d[component_type] = [encoder(component)
                                 for component in
                                 getattr(model, component_type)]

        return d



"""Example of using an initial condition with a fixed amount.

With the amount of free F fixed, most of the free A will be able to bind and
form the A-F complex before equilibrium is reached. If free F were not fixed,
its lower initial amount would severely restrict how much complex could form.
See run_fixed_initial.py for a demonstration of this behavior.

"""

from __future__ import print_function
from pysb import *

Model()

Monomer('A', ['b'])
Monomer('F', ['b'])

Parameter('kf', 1.0)
Parameter('kr', 1.0)
Parameter('A_0', 100)
Parameter('F_0', 20)

Rule('A_bind_F', A(b=None) + F(b=None) | A(b=1) % F(b=1), kf, kr)

Initial(A(b=None), A_0)
Initial(F(b=None), F_0, fixed=True)

Observable('A_free', A(b=None))
Observable('F_free', F(b=None))
Observable('AF_complex', A(b=1) % F(b=1))


if __name__ == '__main__':
    print(__doc__, "\n", model)
    print("""
NOTE: This model code is designed to be imported and programatically
manipulated, not executed directly. The above output is merely a
diagnostic aid.""")



import pickle
from pysb.simulator import SimulationResult
from pysb.simulator.base import _patch_model_setstate
from pysb.importers.sbml import model_from_biomodels

## TO Pickle:
##


def test_unpickle():
    with open('earm_1_0.pkl', 'rb') as f:
        with _patch_model_setstate():
            pickle.load(f)

def test_unpickle_pysb_1_6():
    with open('earm_1_0_pysb1.6.pkl', 'rb') as f:
        with _patch_model_setstate():
            pickle.load(f)

def test_load_legacy_h5():
    SimulationResult.load('earm_1_0_scipyodesimulator.h5')


def test_load_legacy_h5_pysb_1_6():
    SimulationResult.load('earm_1_0_scipyodesimulator_pysb1.6.h5')

def test_import_albeck_biomodels():
    model_from_biomodels('BIOMD0000000220')



from pysb.simulator.base import Simulator, SimulationResult
import scipy.integrate, scipy.sparse
from sympy.utilities.autowrap import CythonCodeWrapper
from sympy.utilities.codegen import (
    C99CodeGen, Routine, InputArgument, OutputArgument, default_datatypes
)
import distutils
import pysb.bng
import sympy
import re
from functools import partial
import numpy as np
import warnings
import os
import inspect
from pysb.logging import get_logger, PySBModelLoggerAdapter, EXTENDED_DEBUG
import logging
import contextlib
import importlib
import tempfile
import shutil
from concurrent.futures import ProcessPoolExecutor, Executor, Future


class ScipyOdeSimulator(Simulator):
    """
    Simulate a model using SciPy ODE integration

    Uses :func:`scipy.integrate.odeint` for the ``lsoda`` integrator,
    :func:`scipy.integrate.ode` for all other integrators.

    .. warning::
        The interface for this class is considered experimental and may
        change without warning as PySB is updated.

    Parameters
    ----------
    model : pysb.Model
        Model to simulate.
    tspan : vector-like, optional
        Time values over which to simulate. The first and last values define
        the time range. Returned trajectories are sampled at every value unless
        the simulation is interrupted for some reason, e.g., due to
        satisfaction of a logical stopping criterion (see 'tout' below).
    initials : vector-like or dict, optional
        Values to use for the initial condition of all species. Ordering is
        determined by the order of model.species. If not specified, initial
        conditions will be taken from model.initials (with initial condition
        parameter values taken from `param_values` if specified).
    param_values : vector-like or dict, optional
        Values to use for every parameter in the model. Ordering is
        determined by the order of model.parameters.
        If passed as a dictionary, keys must be parameter names.
        If not specified, parameter values will be taken directly from
        model.parameters.
    verbose : bool or int, optional (default: False)
        Sets the verbosity level of the logger. See the logging levels and
        constants from Python's logging module for interpretation of integer
        values. False is equal to the PySB default level (currently WARNING),
        True is equal to DEBUG.
    **kwargs : dict
        Extra keyword arguments, including:

        * ``integrator``: Choice of integrator, including ``vode`` (default),
          ``zvode``, ``lsoda``, ``dopri5`` and ``dop853``. See
          :func:`scipy.integrate.ode` for further information.
        * ``integrator_options``: A dictionary of keyword arguments to
          supply to the integrator. See :func:`scipy.integrate.ode`.
        * ``compiler``: Choice of compiler for ODE system: ``cython``,
          or ``python``. Leave unspecified or equal to None for auto-select
          (tries cython, then python). Cython compiles the
          equation system into C code. Python is the slowest but most
          compatible.
        * ``cleanup``: Boolean, whether to delete temporary files.

    Notes
    -----
    If ``tspan`` is not defined, it may be defined in the call to the
    ``run`` method.

    Examples
    --------
    Simulate a model and display the results for an observable:

    >>> from pysb.examples.robertson import model
    >>> import numpy as np
    >>> np.set_printoptions(precision=4)
    >>> sim = ScipyOdeSimulator(model, tspan=np.linspace(0, 40, 10))
    >>> simulation_result = sim.run()
    >>> print(simulation_result.observables['A_total']) \
        #doctest: +NORMALIZE_WHITESPACE
    [1.      0.899   0.8506  0.8179  0.793   0.7728  0.7557  0.7408  0.7277
    0.7158]

    For further information on retrieving trajectories (species,
    observables, expressions over time) from the ``simulation_result``
    object returned by :func:`run`, see the examples under the
    :class:`SimulationResult` class.
    """

    _supports = {'multi_initials': True,
                 'multi_param_values': True}

    # some sane default options for a few well-known integrators
    default_integrator_options = {
        'vode': {
            'method': 'bdf',
            'with_jacobian': True,
            # Set nsteps as high as possible to give our users flexibility in
            # choosing their time step. (Let's be safe and assume vode was
            # compiled with 32-bit ints. What would actually happen if it was
            # and we passed 2**64-1 though?)
            'nsteps': 2 ** 31 - 1,
        },
        'cvode': {
            'method': 'bdf',
            'iteration': 'newton',
        },
        'lsoda': {
            'mxstep': 2**31-1,
        }
    }

    def __init__(self, model, tspan=None, initials=None, param_values=None,
                 verbose=False, **kwargs):

        super(ScipyOdeSimulator, self).__init__(model,
                                                tspan=tspan,
                                                initials=initials,
                                                param_values=param_values,
                                                verbose=verbose,
                                                **kwargs)
        cleanup = kwargs.pop('cleanup', True)
        with_jacobian = kwargs.pop('use_analytic_jacobian', False)
        integrator = kwargs.pop('integrator', 'vode')
        compiler = kwargs.pop('compiler', None)
        integrator_options = kwargs.pop('integrator_options', {})

        if kwargs:
            raise ValueError('Unknown keyword argument(s): {}'.format(
                ', '.join(kwargs.keys())
            ))
        # Generate the equations for the model
        pysb.bng.generate_equations(self._model, cleanup, self.verbose)

        builder_cls = _select_rhs_builder(compiler, self._logger)
        self._logger.debug("Using RhsBuilder: %s", builder_cls.__name__)
        self.rhs_builder = builder_cls(
            model, with_jacobian, cleanup=cleanup, _logger=self._logger
        )

        # build integrator options list from our defaults and any kwargs
        # passed to this function
        options = {}
        if self.default_integrator_options.get(integrator):
            options.update(
                self.default_integrator_options[integrator])  # default options

        options.update(integrator_options)  # overwrite
        # defaults
        self.opts = options

        if integrator != 'lsoda':
            # Only used to check the user has selected a valid integrator
            self.integrator = scipy.integrate.ode(None)
            with warnings.catch_warnings():
                warnings.filterwarnings('error', 'No integrator name match')
                self.integrator.set_integrator(integrator, **options)

    @property
    def _patch_distutils_logging(self):
        """Return distutils logging context manager based on our logger."""
        return _patch_distutils_logging(self._logger.logger)

    def run(self, tspan=None, initials=None, param_values=None,
            num_processors=1):
        """
        Run a simulation and returns the result (trajectories)

        .. note::
            In early versions of the Simulator class, ``tspan``, ``initials``
            and ``param_values`` supplied to this method persisted to future
            :func:`run` calls. This is no longer the case.

        Parameters
        ----------
        tspan
        initials
        param_values
            See parameter definitions in :class:`ScipyOdeSimulator`.
        num_processors : int
            Number of processes to use (default: 1). Set to a larger number
            (e.g. the number of CPU cores available) for parallel execution of
            simulations. This is only useful when simulating with more than one
            set of initial conditions and/or parameters.

        Returns
        -------
        A :class:`SimulationResult` object
        """
        super(ScipyOdeSimulator, self).run(tspan=tspan,
                                           initials=initials,
                                           param_values=param_values,
                                           _run_kwargs=[])
        n_sims = len(self.param_values)

        if num_processors == 1:
            self._logger.debug('Single processor (serial) mode')
        else:
            self._logger.debug('Multi-processor (parallel) mode using {} '
                               'processes'.format(num_processors))

        with SerialExecutor() if num_processors == 1 else \
                ProcessPoolExecutor(max_workers=num_processors) as executor:
            sim_partial = partial(
                _integrator_process,
                tspan=self.tspan,
                integrator_name=self._init_kwargs.get('integrator', 'vode'),
                integrator_opts=self.opts,
                rhs_builder=self.rhs_builder,
            )
            results = [executor.submit(sim_partial, *args)
                       for args in zip(self.initials, self.param_values)]
            try:
                trajectories = [r.result() for r in results]
            finally:
                for r in results:
                    r.cancel()

        tout = np.array([self.tspan] * n_sims)
        self._logger.info('All simulation(s) complete')
        return SimulationResult(self, tout, trajectories)


@contextlib.contextmanager
def _patch_distutils_logging(base_logger):
    """Patch distutils logging functionality with logging.Logger calls.

    The value of the 'base_logger' argument should be a logging.Logger instance,
    and its effective level will be passed on to the patched distutils loggers.

    distutils.log contains its own internal PEP 282 style logging system that
    sends messages straight to stdout/stderr, and numpy.distutils.log extends
    that. This code patches all of this with calls to logging.LoggerAdapter
    instances, and disables the module-level threshold-setting functions so we
    can retain full control over the threshold. Also all WARNING messages are
    "downgraded" to INFO to suppress excessive use of WARNING-level logging in
    numpy.distutils.

    """
    logger = get_logger(__name__)
    logger.debug('patching distutils and numpy.distutils logging')
    logger_methods = 'log', 'debug', 'info', 'warn', 'error', 'fatal'
    other_functions = 'set_threshold', 'set_verbosity'
    saved_symbols = {}
    for module_name in 'distutils.log', 'numpy.distutils.log':
        new_logger = _DistutilsProxyLoggerAdapter(
            base_logger, {'module': module_name}
        )
        module = importlib.import_module(module_name)
        # Save the old values.
        for name in logger_methods + other_functions:
            saved_symbols[module, name] = getattr(module, name)
        # Replace logging functions with bound methods of the Logger object.
        for name in logger_methods:
            setattr(module, name, getattr(new_logger, name))
        # Replace threshold-setting functions with no-ops.
        for name in other_functions:
            setattr(module, name, lambda *args, **kwargs: None)
    try:
        yield
    finally:
        logger.debug('restoring distutils and numpy.distutils logging')
        # Restore everything we overwrote.
        for (module, name), value in saved_symbols.items():
            setattr(module, name, value)


@contextlib.contextmanager
def _set_cflags_no_warnings(logger):
    """ Suppress cython warnings by setting -w flag """
    del_cflags = False
    if 'CFLAGS' not in os.environ \
            and not logger.isEnabledFor(EXTENDED_DEBUG):
        del_cflags = True
        os.environ['CFLAGS'] = '-w'
    try:
        yield
    finally:
        if del_cflags:
            del os.environ['CFLAGS']


class _DistutilsProxyLoggerAdapter(logging.LoggerAdapter):
    """A logging adapter for the distutils logging patcher."""
    def process(self, msg, kwargs):
        return '(from %s) %s' % (self.extra['module'], msg), kwargs
    # Map 'warn' to 'info' to reduce chattiness.
    warn = logging.LoggerAdapter.info
    # Provide 'fatal' to match up with distutils log functions.
    fatal = logging.LoggerAdapter.critical


def _integrator_process(
    initials, param_values, tspan, integrator_name, integrator_opts, rhs_builder
):
    """ Single integrator process, for parallel execution """
    rhs_fn = rhs_builder.rhs_fn
    jac_fn = rhs_builder.jacobian_fn
    initials = np.array(initials, float)
    p = np.array(param_values, float)
    e = rhs_builder.calc_expressions_constant(param_values)
    extra_args = (p, e)

    # LSODA
    if integrator_name == 'lsoda':
        return scipy.integrate.odeint(
            rhs_fn,
            initials,
            tspan,
            args=extra_args,
            Dfun=jac_fn,
            tfirst=True,
            **integrator_opts
        )

    # All other integrators
    integrator = scipy.integrate.ode(rhs_fn, jac=jac_fn)
    with warnings.catch_warnings():
        warnings.filterwarnings('error', 'No integrator name match')
        integrator.set_integrator(integrator_name, **integrator_opts)
    integrator.set_initial_value(initials, tspan[0])

    # Set parameter vectors for RHS func and Jacobian
    integrator.set_f_params(*extra_args)
    if rhs_builder.with_jacobian:
        integrator.set_jac_params(*extra_args)

    trajectory = np.ndarray((len(tspan), rhs_builder.num_species))
    trajectory[0] = initials
    i = 1
    while integrator.successful() and integrator.t < tspan[-1]:
        trajectory[i] = integrator.integrate(tspan[i])
        i += 1
    if integrator.t < tspan[-1]:
        trajectory[i:, :] = 'nan'

    return trajectory


class RhsBuilder:
    """Provides the ODE right-hand side evaluation function, given a model.

    Also provides the Jacobian of the RHS function (partial derivatives of
    species concentrations with respect to all other species).

    This is an abstract base class; concrete subclasses must implement the
    `_get_rhs` and `_get_jacobian` methods. These methods shall return function
    objects for evaluating the RHS and Jacobian, respectively, of the system of
    ODEs. Subclasses may also implement the `check` classmethod which should
    return a bool indicating whether it is usable in the current environment
    (it could for example check that all dependencies are present and working).

    All implementations of this class must be picklable so they can be sent
    across to the ProcessPoolExecutor workers in separate processes. The
    `_get_rhs` and `_get_jacobian` methods may return process-specific
    non-picklable values.

    Parameters
    ----------
    model : pysb.Model
        Model to simulate.
    with_jacobian : bool, optional (default: False)
        Whether to construct the Jacobian evaluation function.
    cleanup : bool, optional (default: True)
        Whether to delete the work directory with generated code and compiled
        output files upon destruction.

    Attributes
    ----------
    with_jacobian : bool
        Whether to construct the Jacobian evaluation function.
    cleanup : bool
        Whether to delete the work directory with generated code and compiled
        output files upon destruction.
    y : sympy.MatrixSymbol
        Symbol "y" representing the amount of all model species.
    p : sympy.MatrixSymbol
        Symbol "p" representing all model parameters.
    e : sympy.MatrixSymbol
        Symbol "e" representing all model constant expressions.
    o : sympy.MatrixSymbol
        Symbol "o" representing all model observables.
    kinetics : sympy.Matrix
        Symbolic form of model reaction rates in terms of y, p, e, o, and time.
    kinetics_jacobian_y : sympy.SparseMatrix
        Symbolic form of partial derivatives of kinetics with respect to y
        (Only set if with_jacobian is True)
    kinetics_jacobian_o : sympy.SparseMatrix
        Symbolic form of partial derivatives of kinetics with respect to o
        (Only set if with_jacobian is True)
    stoichiometry_matrix : scipy.sparse.csr_matrix
        The model's stoichiometry matrix.
    observables_matrix : scipy.sparse.csr_matrix
        Encodes the linear combinations of species that define the model's
        observables.
    model_name : str
        The name of the model.

    """

    def __init__(self, model, with_jacobian=False, cleanup=True, _logger=None):
        self.with_jacobian = with_jacobian
        self.cleanup = cleanup
        self._logger = _logger
        expr_dynamic = model.expressions_dynamic(include_derived=True)
        expr_constant = model.expressions_constant(include_derived=True)
        self.y = sympy.MatrixSymbol('y', len(model.species), 1)
        self.p = sympy.MatrixSymbol('p', len(model.parameters_all()), 1)
        self.e = sympy.MatrixSymbol('e', len(expr_constant), 1)
        self.o = sympy.MatrixSymbol('o', len(model.observables), 1)
        # Parameters symbols. We also need this independently of all_subs.
        param_subs = dict(zip(model.parameters_all(), self.p))
        # All substitution rules we need to apply to the reaction expressions.
        all_subs = param_subs.copy()
        # Species symbols.
        all_subs.update({
            sympy.Symbol('__s%d' % i): self.y[i]
            for i in range(len(model.species))
        })
        # Observables symbols.
        all_subs.update(dict(zip(model.observables, self.o)))
        # Constant expressions symbols.
        all_subs.update(dict(zip(expr_constant, self.e)))
        # Dynamic expressions (expanded).
        all_subs.update({e: e.expand_expr() for e in expr_dynamic})
        self.kinetics = sympy.Matrix([
            r['rate'].subs(all_subs) for r in model.reactions
        ])
        if with_jacobian:
            # The Jacobian can be quite large but it's extremely sparse. We can
            # obtain a sparse representation by making a SparseMatrix copy of
            # the kinetics vector and computing the Jacobian on that.
            kinetics_sparse = sympy.SparseMatrix(self.kinetics)
            # Rather than substitute all the observables linear-combination-of-
            # species expressions into the kinetics and then compute the
            # Jacobian, we can use the total derivative / chain rule to
            # simplify things. The final jacobian of the kinetics must be
            # computed as jac(y) + jac(o) * observables_matrix.
            self._logger.debug("Computing Jacobian matrices")
            self.kinetics_jacobian_y = kinetics_sparse.jacobian(self.y)
            self.kinetics_jacobian_o = kinetics_sparse.jacobian(self.o)
        obs_matrix = scipy.sparse.lil_matrix(
            (len(model.observables), len(model.species)), dtype=np.int64
        )
        for i, obs in enumerate(model.observables):
            obs_matrix[i, obs.species] = obs.coefficients
        self.observables_matrix = obs_matrix.tocsr()
        self.stoichiometry_matrix = model.stoichiometry_matrix
        self.model_name = model.name
        self._expressions_constant = [
            e.expand_expr().xreplace(param_subs) for e in expr_constant
        ]
        self._work_path = None
        self._expressions_constant_fn = None
        self._rhs_fn = None
        self._jacobian_fn = None

    def __del__(self):
        if self._work_path:
            if self.cleanup:
                shutil.rmtree(self._work_path, ignore_errors=True)
                self._logger.debug("Removed work dir: %s", self._work_path)
            else:
                self._logger.debug(
                    "Leaving work dir in place (cleanup=False): %s",
                    self._work_path
                )

    def __getstate__(self):
        state = self.__dict__.copy()
        # Clear cached function objects (will be regenerated on demand).
        state["_expressions_constant_fn"] = None
        state["_rhs_fn"] = None
        state["_jacobian_fn"] = None
        # Loggers are not picklable in Python 3.6.
        state["_logger_extra"] = state["_logger"].extra
        del state["_logger"]
        return state

    def __setstate__(self, state):
        state["_logger"] = PySBModelLoggerAdapter(
            get_logger(self.__module__), state["_logger_extra"]
        )
        del state["_logger_extra"]
        self.__dict__.update(state)

    @property
    def work_path(self):
        """Location for saving any temporary working files."""
        if self._work_path is None:
            # Only initialize this on demand since it will create a directory
            # which requires cleanup on destruction.
            self._work_path = tempfile.mkdtemp(suffix="_pysb_compile")
            self._logger.debug("Created work dir: %s", self._work_path)
        return self._work_path

    @property
    def num_species(self):
        """Number of species in the model."""
        return self.stoichiometry_matrix.shape[0]

    @property
    def num_reactions(self):
        """Number of reactions in the model."""
        return self.stoichiometry_matrix.shape[1]

    @property
    def num_observables(self):
        """Number of observables in the model."""
        return self.observables_matrix.shape[0]

    @property
    def rhs_fn(self):
        """The RHS function rhs(time, y, params, const_exprs)."""
        if self._rhs_fn is None:
            self._logger.debug("Constructing rhs function")
            self._rhs_fn = self._get_rhs()
        return self._rhs_fn

    @property
    def jacobian_fn(self):
        """The Jacobian function jac(time, y, params, const_exprs).

        The value is None if with_jacobian is False."""
        if self.with_jacobian and self._jacobian_fn is None:
            self._logger.debug("Constructing jacobian function")
            self._jacobian_fn = self._get_jacobian()
        return self._jacobian_fn

    def calc_expressions_constant(self, p):
        """Compute constant expressions vector e from parameters vector p."""
        if self._expressions_constant_fn is None:
            # This function is expected to be called rarely enough (once per
            # simulation) that we'll just create it here with lambdify and not
            # allow implementations to override it.
            self._logger.debug("Constructing constant expressions function")
            self._expressions_constant_fn = sympy.lambdify(
                [self.p], self._expressions_constant
            )
        e = self._expressions_constant_fn(p[:, None])
        return np.array(e, float)[:, None]

    def _get_rhs(self):
        """Return the RHS function rhs(time, y, params, const_exprs).

        All subclasses must implement this method."""
        raise NotImplementedError

    def _get_jacobian(self):
        """Return the Jacobian function jac(time, y, params, const_exprs).

        Subclasses may implement this method for Jacobian support."""
        raise NotImplementedError

    @classmethod
    def check(cls):
        """Raises an exception if this class is not usable.

        The default implementation does nothing (always succeeds)."""
        return True

    @classmethod
    def check_safe(cls):
        """Returns a boolean indicating whether this class is usable."""
        try:
            cls.check()
            return True
        except Exception:
            return False


class PythonRhsBuilder(RhsBuilder):

    def _get_rhs(self):
        kinetics = sympy.lambdify(
            [self.y, self.p, self.e, self.o], self.kinetics
        )

        def rhs(t, y, p, e):
            o = (self.observables_matrix * y)[:, None]
            v = kinetics(y[:, None], p[:, None], e, o)
            ydot = self.stoichiometry_matrix * v[:, 0]
            return ydot

        return rhs

    def _get_jacobian(self):
        kinetics_jacobian_y = sympy.lambdify(
            [self.y, self.p, self.e, self.o], self.kinetics_jacobian_y
        )
        kinetics_jacobian_o = sympy.lambdify(
            [self.y, self.p, self.e, self.o], self.kinetics_jacobian_o
        )

        def jacobian(t, y, p, e):
            o = (self.observables_matrix * y)[:, None]
            dy = kinetics_jacobian_y(y[:, None], p[:, None], e, o)
            do = kinetics_jacobian_o(y[:, None], p[:, None], e, o)
            # Compute Jacobian of the kinetics vector by total derivative.
            jv = dy + do * self.observables_matrix
            jac = (self.stoichiometry_matrix * jv).todense()
            return jac

        return jacobian


def _mat_sym_dims(symbol):
    """Return codegen Argument dimensions for a MatrixSymbol."""
    return ((0, symbol.shape[0] - 1), (0, symbol.shape[1] - 1))


class CythonRhsBuilder(RhsBuilder):

    def __init__(self, model, with_jacobian=False, cleanup=True, _logger=None):
        super(CythonRhsBuilder, self).__init__(
            model, with_jacobian, cleanup, _logger
        )
        routine_names = ["kinetics"]
        if with_jacobian:
            routine_names += ["kinetics_jacobian_y", "kinetics_jacobian_o"]
        # We want more control over various details of code generation and
        # wrapper module creation than sympy's autowrap provides, so we'll use
        # the lower-level building blocks directly.
        routines = {name: self._build_routine(name) for name in routine_names}
        code_gen = C99CodeGen()
        extra_compile_args = [
            # The RHS evaluation is such a tiny part of overall integration
            # time, even for huge models, that compiler optimization actually
            # takes more time than it will ever yield back. Since Cython sets
            # -O2 by default we need to override it.
            "-O0",
        ]
        # Opt in to the newer numpy C API which is only supported in Cython 3+.
        import Cython
        if not Cython.__version__.startswith("0."):
            extra_compile_args.append(
                "-DNPY_NO_DEPRECATED_API=NPY_1_7_API_VERSION"
            )
        code_wrapper = CythonCodeWrapper(
            code_gen,
            filepath=self.work_path,
            extra_compile_args=extra_compile_args,
        )
        # Build a module-name-safe string that identifies the model as uniquely
        # as possible to assist in debugging.
        escaped_name = re.sub(
            r"[^A-Za-z0-9]",
            "_",
            self.model_name.encode("unicode_escape").decode(),
        )
        base_name = "pysb_" + escaped_name + "_kinetics"
        code_wrapper._filename = base_name
        code_wrapper._module_basename = base_name + "_wrapper"
        self._logger.debug("Running code generation and Cython compilation")
        functions = {
            name: code_wrapper.wrap_code(routine)
            for name, routine in routines.items()
        }
        # Grab specs for the Cython-compiled modules for reloading later.
        self.module_specs = {
            name: inspect.getmodule(function).__spec__
            for name, function in functions.items()
        }

    def _build_routine(self, name):
        expr = getattr(self, name)
        out = sympy.MatrixSymbol("out", *expr.shape)
        routine = Routine(
            name,
            [
                InputArgument(self.y, dimensions=_mat_sym_dims(self.y)),
                InputArgument(self.p, dimensions=_mat_sym_dims(self.p)),
                InputArgument(self.e, dimensions=_mat_sym_dims(self.e)),
                InputArgument(self.o, dimensions=_mat_sym_dims(self.o)),
                OutputArgument(
                    out,
                    out,
                    expr,
                    datatype=default_datatypes["float"],
                    dimensions=_mat_sym_dims(out),
                ),
            ],
            [],
            [],
            [],
        )
        return routine

    def _load_function(self, name):
        # Import one of our previously compiled Cython functions.
        spec = self.module_specs[name]
        # These two lines seem strangely redundant, but it's how specs work...
        module = spec.loader.create_module(spec)
        spec.loader.exec_module(module)
        function = getattr(module, name + "_c")
        return function

    # The inner functions below are, admittedly, basically identical to those
    # in PythonRhsBuilder. Initial implementations differed more but they
    # converged over time. They have not been refactored out into a common
    # implementation to allow for exploration of performance improvements.

    def _get_rhs(self):
        kinetics = self._load_function("kinetics")

        def rhs(t, y, p, e):
            o = (self.observables_matrix * y)[:, None]
            v = kinetics(y[:, None], p[:, None], e, o)
            ydot = self.stoichiometry_matrix * v[:, 0]
            return ydot

        return rhs

    def _get_jacobian(self):
        kinetics_jacobian_y = self._load_function("kinetics_jacobian_y")
        kinetics_jacobian_o = self._load_function("kinetics_jacobian_o")

        def jacobian(t, y, p, e):
            o = (self.observables_matrix * y)[:, None]
            dy = kinetics_jacobian_y(y[:, None], p[:, None], e, o)
            do = kinetics_jacobian_o(y[:, None], p[:, None], e, o)
            # Compute Jacobian of the kinetics vector by total derivative.
            jv = dy + do * self.observables_matrix
            jac = self.stoichiometry_matrix * jv
            return jac

        return jacobian

    _check_ok = False

    @classmethod
    def check(cls):
        # Quick return if previous check succeeded, otherwise re-run the checks
        # so the exception is always raised if called repeatedly.
        if cls._check_ok:
            return
        compiler_exc = None
        try:
            import Cython
            Cython.inline("x = 1", force=True, quiet=True)
            cls._check_ok = True
            return
        except ImportError:
            raise RuntimeError(
                "Please install the Cython package to use this compiler."
            )
        # Catch some common C compiler configuration problems so we can raise a
        # chained exception with a more helpful message.
        except (
            Cython.Compiler.Errors.CompileError,
            distutils.errors.CCompilerError,
            distutils.errors.DistutilsPlatformError,
        ) as e:
            compiler_exc = e
        except ValueError as e:
            if e.args == ("Symbol table not found",):
                # This is a common error building against numpy with mingw32.
                compiler_exc = e
            else:
                raise
        message = "Please check you have a functional C compiler"
        if os.name == "nt":
            message += ", available from " \
                       "https://wiki.python.org/moin/WindowsCompilers"
        else:
            message += "."
        # Reference chained exceptions for specifics on what went wrong.
        message += "\nSee the above error messages for more details."
        raise RuntimeError(message) from compiler_exc


# Keep these sorted in priority order.
_rhs_builders = {
    "cython": CythonRhsBuilder,
    "python": PythonRhsBuilder,
}


def _select_rhs_builder(compiler, logger):
    if compiler is None:
        # Probe for the first (best) working builder.
        for cls in _rhs_builders.values():
            if cls.check_safe():
                break
        else:
            raise RuntimeError("No usable ODE compiler found.")
        if cls is PythonRhsBuilder:
            logger.warning(
                "This system of ODEs will be evaluated in pure Python. "
                "This may be slow for large models. We recommend "
                "installing the 'cython' package for compiling the ODEs to "
                "C code. This warning can be suppressed by specifying "
                "compiler='python'."
            )
    else:
        try:
            cls = _rhs_builders[compiler]
        except KeyError:
            raise ValueError('Unknown ODE compiler name: %s' % compiler) \
                from None
        cls.check()
    return cls


class SerialExecutor(Executor):
    """ Execute tasks in serial (immediately on submission) """
    def submit(self, fn, *args, **kwargs):
        f = Future()
        try:
            result = fn(*args, **kwargs)
        except BaseException as e:
            f.set_exception(e)
        else:
            f.set_result(result)

        return f



from pysb.builder import Builder
try:
    import trum
    from truml.readers import KappaReader
    import pyparsing as pp
except ImportError:
    KappaReader = None
    pp = None
import io

class KappaImportError(Exception):
    pass


class KappaBuilder(Builder):
    """
    Assemble a Model from a .ka file.

    See :py:func:`model_from_kappa` for further details.
    """
    def __init__(self, filename, force=False, cleanup=True):
        super(KappaBuilder, self).__init__()
        if KappaReader is None:
            raise ImportError('TRuML is needed (e.g. pip install truml)')
        kr = KappaReader(filename)
        try:
            model = kr.parse()
        except pp.ParseException as pe:
            raise KappaImportError(pe)

        out = io.StringIO
        model.write_as_bngl(out, False)
        print(out)
        raise


def model_from_kappa(filename, force=False, cleanup=True):
    """
    Convert a BioNetGen .bngl model file into a PySB Model.

    Notes
    -----

    The following features are not supported in PySB and will cause an error
    if present in a .bngl file:

    * Fixed species (with a ``$`` prefix, like ``$Null``)
    * BNG excluded or included reaction patterns (deprecated in BNG)
    * BNG local functions
    * Molecules with identically named sites, such as ``M(l,l)``
    * BNG's custom rate law functions, such as ``MM`` and ``Sat``
      (deprecated in BNG)

    Parameters
    ----------
    filename : string
        A BioNetGen .bngl file
    force : bool, optional
        The default, False, will raise an Exception if there are any errors
        importing the model to PySB, e.g. due to unsupported features.
        Setting to True will attempt to ignore any import errors, which may
        lead to a model that only poorly represents the original. Use at own
        risk!
    cleanup : bool
        Delete temporary directory on completion if True. Set to False for
        debugging purposes.
    """
    bb = KappaBuilder(filename, force=force, cleanup=cleanup)
    return bb.model



"""
Module containing a class to return the StochKit XML equivalent of a model

Contains code based on the `gillespy <https://github.com/JohnAbel/gillespy>`
library with permission from author Brian Drawert.

For information on how to use the model exporters, see the documentation
for :py:mod:`pysb.export`.
"""
from pysb.export import Exporter, CompartmentsNotSupported
from pysb.core import as_complex_pattern, Expression, Parameter
from pysb.bng import generate_equations
import numpy as np
import sympy
import re
from collections import defaultdict
import itertools
try:
    import lxml.etree as etree
    pretty_print = True
except ImportError:
    import xml.etree.ElementTree as etree
    import xml.dom.minidom
    pretty_print = False


class StochKitExporter(Exporter):
    """A class for returning the Kappa for a given PySB model.

    Inherits from :py:class:`pysb.export.Exporter`, which implements
    basic functionality for all exporters.
    """
    @staticmethod
    def _species_to_element(species_num, species_val):
        e = etree.Element('Species')
        idElement = etree.Element('Id')
        idElement.text = species_num
        e.append(idElement)

        initialPopulationElement = etree.Element('InitialPopulation')
        initialPopulationElement.text = str(species_val)
        e.append(initialPopulationElement)

        return e


    @staticmethod
    def _parameter_to_element(param_name, param_val):
        e = etree.Element('Parameter')
        idElement = etree.Element('Id')
        idElement.text = param_name
        e.append(idElement)
        expressionElement = etree.Element('Expression')
        expressionElement.text = str(param_val)
        e.append(expressionElement)
        return e


    @staticmethod
    def _reaction_to_element(rxn_name, rxn_desc, propensity_fxn, reactants,
                             products):
        e = etree.Element('Reaction')

        idElement = etree.Element('Id')
        idElement.text = rxn_name
        e.append(idElement)

        descriptionElement = etree.Element('Description')
        descriptionElement.text = rxn_desc
        e.append(descriptionElement)

        typeElement = etree.Element('Type')
        if isinstance(propensity_fxn, (Parameter, float)):
            typeElement.text = 'mass-action'
            e.append(typeElement)
            rateElement = etree.Element('Rate')
            rateElement.text = propensity_fxn.name if isinstance(
                propensity_fxn, Parameter) else str(propensity_fxn)
            e.append(rateElement)
        else:
            typeElement.text = 'customized'
            e.append(typeElement)
            functionElement = etree.Element('PropensityFunction')
            functionElement.text = propensity_fxn
            e.append(functionElement)

        reactantElement = etree.Element('Reactants')

        for reactant, stoichiometry in reactants.items():
            srElement = etree.Element('SpeciesReference')
            srElement.set('id', reactant)
            srElement.set('stoichiometry', str(stoichiometry))
            reactantElement.append(srElement)

        e.append(reactantElement)

        productElement = etree.Element('Products')
        for product, stoichiometry in products.items():
            srElement = etree.Element('SpeciesReference')
            srElement.set('id', product)
            srElement.set('stoichiometry', str(stoichiometry))
            productElement.append(srElement)
        e.append(productElement)

        return e

    def export(self, initials=None, param_values=None):
        """Generate the corresponding StochKit2 XML for a PySB model

        Parameters
        ----------
        initials : list of numbers
            List of initial species concentrations overrides
            (must be same length as model.species). If None,
            the concentrations from the model are used.
        param_values : list
            List of parameter value overrides (must be same length as
            model.parameters). If None, the parameter values from the model
            are used.

        Returns
        -------
        string
            The model in StochKit2 XML format
        """
        if self.model.compartments:
            raise CompartmentsNotSupported()

        generate_equations(self.model)
        document = etree.Element("Model")

        d = etree.Element('Description')

        d.text = 'Exported from PySB Model: %s' % self.model.name
        document.append(d)

        # Number of Reactions
        nr = etree.Element('NumberOfReactions')
        nr.text = str(len(self.model.reactions))
        document.append(nr)

        # Number of Species
        ns = etree.Element('NumberOfSpecies')
        ns.text = str(len(self.model.species))
        document.append(ns)

        if param_values is None:
            # Get parameter values from model if not supplied
            param_values = [p.value for p in self.model.parameters]

        # Add in derived parameters if needed
        if self.model._derived_parameters and len(param_values) == len(
                self.model.parameters):
            param_values += [p.value for p in self.model._derived_parameters]
        elif len(param_values) != len(self.model.parameters) + len(
                self.model._derived_parameters
        ):
            raise ValueError('param_values must be a list of numeric '
                             'parameter values the same length as '
                             'model.parameters, optionally including '
                             'derived parameters on the end (if model contains '
                             'local functions)')

        # Get initial species concentrations from model if not supplied
        if initials is None:
            initials = np.zeros((len(self.model.species),))
            subs = dict((p, param_values[i]) for i, p in
                        enumerate(self.model.parameters))

            for ic in self.model.initials:
                cp = as_complex_pattern(ic.pattern)
                si = self.model.get_species_index(cp)
                if si is None:
                    raise IndexError("Species not found in model: %s" %
                                     repr(cp))
                if ic.value in self.model.parameters:
                    pi = self.model.parameters.index(ic.value)
                    value = param_values[pi]
                elif ic.value in self.model.expressions:
                    value = ic.value.expand_expr().evalf(subs=subs)
                else:
                    raise ValueError(
                        "Unexpected initial condition value type")
                initials[si] = value
        else:
            # Validate length
            if len(initials) != len(self.model.species):
                raise Exception('initials must be a list of numeric initial '
                                'concentrations the same length as '
                                'model.species')

        # Species
        spec = etree.Element('SpeciesList')
        for s_id in range(len(self.model.species)):
            spec.append(self._species_to_element('__s%d' % s_id,
                                                 initials[s_id]))
        document.append(spec)

        # Parameters
        params = etree.Element('ParametersList')
        for p_id, param in enumerate(itertools.chain(
                self.model.parameters, self.model._derived_parameters)):
            p_name = param.name
            if p_name == 'vol':
                p_name = '__vol'
            p_value = param.value if param_values is None else \
                param_values[p_id]
            params.append(self._parameter_to_element(p_name, p_value))
        # Default volume parameter value
        params.append(self._parameter_to_element('vol', 1.0))

        document.append(params)

        # Expressions and observables
        expr_strings = {
            e.name: '(%s)' % sympy.ccode(
                e.expand_expr(expand_observables=True)
            )
            for e in itertools.chain(
                self.model.expressions_constant(),
                self.model.expressions_dynamic(include_local=False),
                self.model._derived_expressions
            )
        }

        # Reactions
        reacs = etree.Element('ReactionsList')
        pattern = re.compile("(__s\d+)\*\*(\d+)")
        for rxn_id, rxn in enumerate(self.model.reactions):
            rxn_name = 'Rxn%d' % rxn_id
            rxn_desc = 'Rules: %s' % str(rxn["rule"])

            reactants = defaultdict(int)
            products = defaultdict(int)
            # reactants
            for r in rxn["reactants"]:
                reactants["__s%d" % r] += 1
            # products
            for p in rxn["products"]:
                products["__s%d" % p] += 1
            # replace terms like __s**2 with __s*(__s-1)
            rate = str(rxn["rate"])

            matches = pattern.findall(rate)
            for m in matches:
                repl = m[0]
                for i in range(1, int(m[1])):
                    repl += "*(%s-%d)" % (m[0], i)
                rate = re.sub(pattern, repl, rate, count=1)

            # expand only expressions used in the rate eqn
            for e in {sym for sym in rxn["rate"].atoms()
                      if isinstance(sym, Expression)}:
                rate = re.sub(r'\b%s\b' % e.name,
                              expr_strings[e.name],
                              rate)

            total_reactants = sum(reactants.values())
            rxn_params = rxn["rate"].atoms(Parameter)
            rate = None
            if total_reactants <= 2 and len(rxn_params) == 1:
                # Try to parse as mass action to avoid compiling custom
                # propensity functions in StochKit (slow for big models)
                rxn_param = rxn_params.pop()
                putative_rate = sympy.Mul(*[sympy.symbols(r) ** r_stoich for
                                            r, r_stoich in
                                            reactants.items()]) * rxn_param

                rxn_floats = rxn["rate"].atoms(sympy.Float)
                rate_mul = 1.0
                if len(rxn_floats) == 1:
                    rate_mul = next(iter(rxn_floats))
                    putative_rate *= rate_mul

                if putative_rate == rxn["rate"]:
                    # Reaction is mass-action, set rate to a Parameter or float
                    if len(rxn_floats) == 0:
                        rate = rxn_param
                    elif len(rxn_floats) == 1:
                        rate = rxn_param.value * float(rate_mul)

                    if rate is not None and len(reactants) == 1 and \
                            max(reactants.values()) == 2:
                        # Need rate * 2 in addition to any rate factor
                        rate = (rate.value if isinstance(rate, Parameter)
                                else rate) * 2.0

            if rate is None:
                # Custom propensity function needed

                if isinstance(rxn['rate'], Expression):
                    rate = expr_strings[rxn['rate'].name]
                else:
                    rxn_atoms = rxn["rate"].atoms()

                    # replace terms like __s**2 with __s*(__s-1)
                    rate = str(rxn["rate"])

                    matches = pattern.findall(rate)
                    for m in matches:
                        repl = m[0]
                        for i in range(1, int(m[1])):
                            repl += "*(%s-%d)" % (m[0], i)
                        rate = re.sub(pattern, repl, rate, count=1)

                    # expand only expressions used in the rate eqn
                    for e in {sym for sym in rxn_atoms
                              if isinstance(sym, Expression)}:
                        rate = re.sub(r'\b%s\b' % e.name,
                                      expr_strings[e.name],
                                      rate)

            reacs.append(self._reaction_to_element(rxn_name,
                                                   rxn_desc,
                                                   rate,
                                                   reactants,
                                                   products))
        document.append(reacs)

        if pretty_print:
            return etree.tostring(document, pretty_print=True).decode('utf8')
        else:
            # Hack to print pretty xml without pretty-print
            # (requires the lxml module).
            doc = etree.tostring(document)
            xmldoc = xml.dom.minidom.parseString(doc)
            uglyXml = xmldoc.toprettyxml(indent='  ')
            text_re = re.compile(">\n\s+([^<>\s].*?)\n\s+</", re.DOTALL)
            prettyXml = text_re.sub(">\g<1></", uglyXml)
            return prettyXml



# -*- coding:utf-8; -*-
"""
Test example models export without error using the available exporters

Based on code submitted in a PR by @keszybz in pysb/pysb#113
"""
from pysb.tests.test_examples import get_example_models, expected_exceptions
from pysb import export
from pysb.export.json import JsonExporter
from pysb.simulator import ScipyOdeSimulator
from pysb.importers.bngl import model_from_bngl
import numpy as np
import pandas as pd
import tempfile
import os
import sys
try:
    import roadrunner
except ImportError:
    roadrunner = None
from nose.plugins.skip import SkipTest
from pysb.importers.json import model_from_json
from pysb.testing import check_model_against_component_list


# Pairs of model, format that are expected to be incompatible.
skip_combinations = {
    ('fixed_initial', 'kappa'),
}


def test_export():
    for model in get_example_models():
        for format in export.formats:
            if (base_name(model), format) in skip_combinations:
                continue
            fn = lambda: check_convert(model, format)
            fn.description = "Check export: {} → {}".format(
                model.name, format)
            yield fn


def base_name(model):
    # This needs to handle names with zero or more dots.
    return model.name.split('.')[-1]


def check_convert(model, format):
    """ Test exporters run without error """
    exported_file = None
    try:
        if format == 'json':
            exported_file = JsonExporter(model).export(include_netgen=True)
        else:
            exported_file = export.export(model, format)
    except export.ExpressionsNotSupported:
        pass
    except export.CompartmentsNotSupported:
        pass
    except export.LocalFunctionsNotSupported:
        pass
    except Exception as e:
        # Some example models are deliberately incomplete, so here we
        # will treat any of these "expected" exceptions as a success.
        exception_class = expected_exceptions.get(base_name(model))
        if not exception_class or not isinstance(e, exception_class):
            raise

    if exported_file is not None:
        if format == 'python':
            # linspace arguments picked to avoid VODE warning
            exec(exported_file + 'Model().simulate(tspan=numpy.linspace(0,1,501))\n', {'numpy': np})
        elif format == 'pysb_flat':
            exec(exported_file, {'__name__': model.name})
        elif format == 'sbml':
            # Skip the simulation comparison if roadrunner not available
            if roadrunner is None:
                raise SkipTest("SBML Simulation test skipped (requires roadrunner)")

            roadrunner.Logger.setLevel(roadrunner.Logger.LOG_ERROR)

            # Simulate SBML using roadrunner
            rr = roadrunner.RoadRunner(exported_file)
            rr.timeCourseSelections = \
                ['__s{}'.format(i) for i in range(len(model.species))] + \
                ['__obs{}'.format(i) for i in range(len(model.observables))]
            rr_result = rr.simulate(0, 10, 100)

            # Simulate original using PySB
            df = ScipyOdeSimulator(model).run(tspan=np.linspace(0, 10, 100)).dataframe

            # Compare species' trajectories
            for sp_idx in range(len(model.species)):
                rr_sp = rr_result[:, sp_idx]
                py_sp = df.iloc[:, sp_idx]
                is_close = np.allclose(rr_sp, py_sp, rtol=1e-4)
                if not is_close:
                    print(pd.DataFrame(dict(rr=rr_sp, pysb=py_sp)))
                    raise ValueError('Model {}, species __s{} trajectories do not match:'.format(
                        model.name, sp_idx))

            # Compare observables' trajectories
            for obs_idx in range(len(model.observables)):
                rr_obs = rr_result[:, obs_idx + len(model.species)]
                py_obs = df.iloc[:, obs_idx + len(model.species)]
                is_close = np.allclose(rr_obs, py_obs, rtol=1e-4)
                if not is_close:
                    print(pd.DataFrame(dict(rr=rr_obs, pysb=py_obs)))
                    raise ValueError('Model {}, observable__o{} "{}" trajectories do not match:'.format(
                        model.name, obs_idx, model.observables[obs_idx].name))
        elif format == 'json':
            # Round-trip the model by re-importing the JSON
            m = model_from_json(exported_file)
            # Check network generation and force RHS evaluation
            if model.name not in ('pysb.examples.tutorial_b',
                                  'pysb.examples.tutorial_c'):
                ScipyOdeSimulator(m, compiler='cython')
                if sys.version_info.major >= 3:
                    # Only check on Python 3 to avoid string-to-unicode encoding
                    # issues
                    check_model_against_component_list(
                        m, model.all_components())
                # Check observable generation
                for obs in model.observables:
                    assert obs.coefficients == \
                        m.observables[obs.name].coefficients
                    assert obs.species == \
                        m.observables[obs.name].species
        elif format == 'bngl':
            if model.name.endswith('tutorial_b') or \
                    model.name.endswith('tutorial_c'):
                # Models have no rules
                return
            with tempfile.NamedTemporaryFile(suffix='.bngl',
                                             delete=False) as tf:
                tf.write(exported_file.encode('utf8'))
                # Cannot have two simultaneous file handled on Windows
                tf.close()

                try:
                    m = model_from_bngl(tf.name)
                    # Generate network and force RHS evaluation
                    ScipyOdeSimulator(m, compiler='cython')
                finally:
                    os.unlink(tf.name)



from __future__ import print_function
from pysb.simulator import ScipyOdeSimulator
from tutorial_a import model

t = [0, 10, 20, 30, 40, 50, 60]
simulator = ScipyOdeSimulator(model, tspan=t)
simresult = simulator.run()
print(simresult.species)



from pysb.examples.tyson_oscillator import model
from pysb.simulator import ScipyOdeSimulator
import numpy as np
import matplotlib.pyplot as plt

t = np.linspace(0, 100, 10001)
x = ScipyOdeSimulator(model).run(tspan=t).all

plt.plot(t, x['CT'],  lw=2, label='CT')  # Good validation of mass balance for cdc2, should be constant at 1
plt.plot(t, x['YT'], lw=2, label='YT')
plt.plot(t, x['M'], lw=2, label='M')

plt.legend(loc=0)
plt.xlabel('time')
plt.ylabel('population')

plt.show()
